from datetime import timedelta
from os import PathLike
from pathlib import Path
from typing import List

from coupledmodeldriver.platforms import Platform
from coupledmodeldriver.script import (
    JobScript,
    EnsembleGenerationJob,
    EnsembleRunScript,
    EnsembleCleanupScript,
    slurm_dependencies,
    slurm_submit_get_id,
    bash_if,
)


class SchismJob(JobScript):
    """
    abstraction of a job script for running SCHISM
    """

    def __init__(
        self,
        platform: Platform,
        commands: List[str],
        slurm_run_name: str = None,
        slurm_tasks: int = None,
        slurm_duration: timedelta = None,
        slurm_account: str = None,
        source_filename: PathLike = None,
        **kwargs,
    ):
        if slurm_run_name is None:
            slurm_run_name = 'SCHISM_JOB'

        super().__init__(
            platform=platform,
            commands=commands,
            slurm_run_name=slurm_run_name,
            slurm_tasks=slurm_tasks,
            slurm_duration=slurm_duration,
            slurm_account=slurm_account,
            **kwargs,
        )

        if source_filename is not None:
            if not isinstance(source_filename, Path):
                source_filename = Path(source_filename)
            source_filename = source_filename.as_posix()
            if len(str(source_filename)) > 0:
                self.commands.insert(0, f'source {source_filename}')


class SchismRunJob(SchismJob):
    """
    job script for running SCHISM
    """

    def __init__(
        self,
        platform: Platform,
        slurm_run_name: str = None,
        slurm_tasks: int = None,
        slurm_duration: timedelta = None,
        slurm_account: str = None,
        executable: PathLike = None,
        commands: List[str] = None,
        num_strides: int = 4,
        old_io: bool = False,
        **kwargs,
    ):
        super().__init__(
            platform=platform,
            commands=commands,
            slurm_run_name=slurm_run_name,
            slurm_tasks=slurm_tasks,
            slurm_duration=slurm_duration,
            slurm_account=slurm_account,
            **kwargs,
        )

        if executable is None:
            executable = 'pschism-TVD_VL'
        else:
            if not isinstance(executable, Path):
                executable = Path(executable)
            executable = executable.as_posix()

        self.executable = executable
        if old_io:
            self.commands.append(
                f'{self.launcher} {self.executable}'
                if self.launcher is not None
                else f'{self.executable}'
            )
        else:
            self.commands.append(
                f'{self.launcher} {self.executable} {num_strides}'
                if self.launcher is not None
                else f'{self.executable} {num_strides}'
            )


class SchismCombineSchoutJob(SchismJob):
    def __init__(
        self,
        platform: Platform,
        start_idx: int,
        end_idx: int,
        slurm_run_name: str = None,
        slurm_tasks: int = None,
        slurm_duration: timedelta = None,
        slurm_account: str = None,
        executable: PathLike = None,
        commands: List[str] = None,
        **kwargs,
    ):
        super().__init__(
            platform=platform,
            commands=commands,
            slurm_run_name=slurm_run_name,
            slurm_tasks=slurm_tasks,
            slurm_duration=slurm_duration,
            slurm_account=slurm_account,
            **kwargs,
        )

        if executable is None:
            executable = 'combine_output11'
        else:
            if not isinstance(executable, Path):
                executable = Path(executable)
            executable = executable.as_posix()

        self.executable = executable
        self.start_idx = start_idx
        self.end_idx = end_idx
        self.commands.extend(
            [
                'pushd outputs',
                f'{self.executable} --begin {self.start_idx} --end {self.end_idx}',
                'popd',
            ]
        )


class SchismCombineHotstartJob(SchismJob):
    def __init__(
        self,
        platform: Platform,
        iterations_idx: List[int],
        slurm_run_name: str = None,
        slurm_tasks: int = None,
        slurm_duration: timedelta = None,
        slurm_account: str = None,
        executable: PathLike = None,
        commands: List[str] = None,
        **kwargs,
    ):
        super().__init__(
            platform=platform,
            commands=commands,
            slurm_run_name=slurm_run_name,
            slurm_tasks=slurm_tasks,
            slurm_duration=slurm_duration,
            slurm_account=slurm_account,
            **kwargs,
        )

        if executable is None:
            executable = 'combine_hotstart7'
        else:
            if not isinstance(executable, Path):
                executable = Path(executable)
            executable = executable.as_posix()

        self.executable = executable
        self.iterations_idx = iterations_idx
        self.commands.extend(
            [
                'pushd outputs',
                *[f'{self.executable} --iteration {it}' for it in self.iterations_idx],
                'popd',
            ]
        )


class SchismEnsembleGenerationJob(EnsembleGenerationJob):
    """
    job script to generate the ensemble configuration
    """

    def __init__(
        self,
        platform: Platform,
        slurm_tasks: int = None,
        slurm_duration: timedelta = None,
        slurm_account: str = None,
        commands: List[str] = None,
        parallel: bool = False,
        **kwargs,
    ):
        super().__init__(
            platform=platform,
            generate_command='generate_schism',
            commands=commands,
            slurm_run_name='SCHISM_GENERATE_CONFIGURATION',
            slurm_tasks=slurm_tasks,
            slurm_duration=slurm_duration,
            slurm_account=slurm_account,
            parallel=parallel,
            **kwargs,
        )


class SchismEnsembleRunScript(EnsembleRunScript):
    """
    script to run the ensemble, either by running it directly or by submitting model execution to the job manager
    default filename is ``run_<platform>.sh``
    """

    def _spinup_lines(self):
        spinup_lines = ['# run spinup', 'pushd ${DIRECTORY}/spinup >/dev/null 2>&1']
        if self.platform.value['uses_slurm']:
            spinup_lines.extend(
                [
                    slurm_submit_get_id('schism.job', 'spinup_jobid'),
                    slurm_submit_get_id(
                        'combine_hotstart.job',
                        'combine_hotstart_jobid',
                        dependencies=slurm_dependencies(after_ok=['$spinup_jobid']),
                    ),
                ]
            )
        else:
            spinup_lines.extend(['sh schism.job', 'sh combine_hotstart.job'])
        spinup_lines.extend(['popd >/dev/null 2>&1', ''])

        return spinup_lines

    def _hotstart_lines(self):
        hotstart_lines = ['pushd ${hotstart} >/dev/null 2>&1']
        if self.platform.value['uses_slurm']:
            after_ok = []
            if self.run_spinup:
                after_ok.append('$combine_hotstart_jobid')
            dependencies = slurm_dependencies(after_ok=after_ok)
            hotstart_lines.extend(
                [
                    slurm_submit_get_id('schism.job', 'run_jobid', dependencies=dependencies),
                    bash_if(
                        '[ -f combine_hotstart.job ]',
                        slurm_submit_get_id(
                            'combine_hotstart.job',
                            '_',
                            dependencies=slurm_dependencies(after_ok=['$run_jobid']),
                        ),
                    ),
                    bash_if(
                        '[ -f combine_output.job ]',
                        slurm_submit_get_id(
                            'combine_output.job',
                            '_',
                            dependencies=slurm_dependencies(after_ok=['$run_jobid']),
                        ),
                    ),
                ]
            )
        else:
            hotstart_lines.extend(
                [
                    'sh schism.job',
                    bash_if('[ -f combine_hotstart.job ]', 'sh combine_hotstart.job &'),
                    bash_if('[ -f combine_output.job ]', 'sh combine_output.job'),
                ]
            )
        hotstart_lines.append('popd >/dev/null 2>&1')

        return hotstart_lines


class SchismEnsembleCleanupScript(EnsembleCleanupScript):
    """
    script for cleaning an ensemble configuration, by deleting output and log files
    """

    def __init__(self, commands: List[str] = None):
        filenames = ['outputs/*']
        spinup_filenames = []
        hotstart_filenames = []

        super().__init__(commands, filenames, spinup_filenames, hotstart_filenames)



from copy import copy
from datetime import datetime, timedelta
from os import PathLike
from pathlib import Path
from typing import Any, Dict, List

from nemspy import ModelingSystem
from nemspy.model.base import ModelEntry
from pyschism.driver import ModelDriver
from pyschism.mesh import Hgrid

from coupledmodeldriver.configure import NEMSJSON
from coupledmodeldriver.configure.base import (
    ConfigurationJSON,
    ModelDriverJSON,
    NEMSCapJSON,
    SlurmJSON,
)
from coupledmodeldriver.configure.configure import RunConfiguration
from coupledmodeldriver.configure.forcings.base import (
    ForcingJSON,
    TidalForcingJSON,
    BestTrackForcingJSON,
    NationalWaterModelFocringJSON,
)
from coupledmodeldriver.generate.schism.base import SCHISMJSON
from coupledmodeldriver.platforms import Platform
from coupledmodeldriver.utilities import LOGGER


class SCHISMRunConfiguration(RunConfiguration):
    """
    run configuration for SCHISM-only run, with optional tidal and / or best track forcing using SCHISM's input capability
    """

    REQUIRED = {
        ModelDriverJSON,
        SlurmJSON,
        SCHISMJSON,
    }
    SUPPLEMENTARY = {
        TidalForcingJSON,
        BestTrackForcingJSON,
        NationalWaterModelFocringJSON,
    }

    def __init__(
        self,
        mesh_directory: PathLike,
        modeled_start_time: datetime,
        modeled_end_time: datetime,
        modeled_timestep: timedelta,
        tidal_spinup_duration: timedelta = None,
        platform: Platform = None,
        perturbations: Dict[str, Dict[str, Any]] = None,
        forcings: List[ForcingJSON] = None,
        schism_processors: int = None,
        slurm_job_duration: timedelta = None,
        slurm_partition: str = None,
        slurm_email_address: str = None,
        schism_executable: PathLike = None,
        schism_hotstart_combiner: PathLike = None,
        schism_schout_combiner: PathLike = None,
        schism_use_old_io: bool = False,
        source_filename: PathLike = None,
    ):
        """
        :param mesh_directory: path to input mesh directory (containing ``fort.13``, ``fort.14``)
        :param modeled_start_time: start time within the modeled system
        :param modeled_end_time: end time within the modeled system
        :param modeled_timestep: time interval within the modeled system
        :param tidal_spinup_duration: spinup time for SCHISM tidal coldstart
        :param platform: HPC platform for which to configure
        :param perturbations: dictionary of runs encompassing run names to parameter values
        :param forcings: list of forcing configurations to connect to SCHISM
        :param schism_processors: numbers of processors to assign for SCHISM
        :param slurm_job_duration: wall clock time of job
        :param slurm_partition: Slurm partition
        :param slurm_email_address: email address to send Slurm notifications
        :param schism_executable: filename of compiled ``pschism-TVD_VL``
        :param schism_hotstart_combiner: filename of compiled hotstart combiner
        :param schism_schout_combiner: filename of compiled SCHISM old output combiner
        :param schism_use_old_io: flag to indicate if the compiled SCHISM uses old IO
        :param source_filename: path to module file to ``source``
        """

        self.__pyschism_mesh = None

        if not isinstance(mesh_directory, Path):
            mesh_directory = Path(mesh_directory)

        if platform is None:
            platform = Platform.LOCAL

        if forcings is None:
            forcings = []

        if schism_processors is None:
            schism_processors = 11

        if schism_executable is None:
            schism_executable = 'pschism-TVD_VL'

        if schism_hotstart_combiner is None:
            schism_hotstart_combiner = 'combine_hotstart7'

        if schism_schout_combiner is None:
            schism_schout_combiner = 'combine_output11'

        slurm = SlurmJSON(
            account=platform.value['slurm_account'],
            tasks=schism_processors,
            partition=slurm_partition,
            job_duration=slurm_job_duration,
            email_address=slurm_email_address,
        )

        schism = SCHISMJSON(
            schism_executable_path=schism_executable,
            schism_hotstart_combiner_path=schism_hotstart_combiner,
            schism_schout_combiner_path=schism_schout_combiner,
            schism_use_old_io=schism_use_old_io,
            modeled_start_time=modeled_start_time,
            modeled_end_time=modeled_end_time,
            modeled_timestep=modeled_timestep,
            fgrid_path=mesh_directory / 'manning.gr3',
            hgrid_path=mesh_directory / 'hgrid.gr3',
            tidal_spinup_duration=tidal_spinup_duration,
            source_filename=source_filename,
            slurm_configuration=slurm,
            processors=schism_processors,
        )

        driver = ModelDriverJSON(platform=platform, perturbations=perturbations)
        super().__init__([driver, slurm, schism])

        for forcing in forcings:
            self.add_forcing(forcing)

    def add_forcing(self, forcing: ForcingJSON):
        if forcing not in self:
            forcing = self[self.add(forcing)]
            try:
                self['schism'].add_forcing(forcing)
            except Exception as error:
                LOGGER.error(error)

    @property
    def forcings(self) -> List[ForcingJSON]:
        return [
            configuration
            for configuration in self.configurations
            if isinstance(configuration, ForcingJSON)
        ]

    @property
    def pyschism_mesh(self) -> Hgrid:
        return self['schism'].pyschism_mesh

    @pyschism_mesh.setter
    def pyschism_mesh(self, pyschism_mesh: Hgrid):
        self['schism'].pyschism_mesh = pyschism_mesh

    @property
    def pyschism_driver(self) -> ModelDriver:
        return self['schism'].pyschism_driver

    def files_exist(self, directory: PathLike) -> bool:
        if not isinstance(directory, Path):
            directory = Path(directory)

        if not directory.exists():
            return False

        files_to_write = [
            'bctides.in',
            'hgrid.gr3',
            'hgrid.ll',
            'manning.gr3',
            'outputs',
            'param.nml',
            'vgrid.in',
            'schism.job',
        ]
        if 'nems' in self:
            files_to_write.extend(
                ['nems.configure', 'atm_namelist.rc', 'model_configure', 'config.rc',]
            )
        existing_files = [filename.name for filename in directory.iterdir()]

        return all([filename in existing_files for filename in files_to_write])

    def __copy__(self) -> 'SCHISMRunConfiguration':
        return self.__class__.from_configurations(
            [copy(configuration) for configuration in self.configurations]
        )

    @classmethod
    def from_configurations(
        cls, configurations: List[ConfigurationJSON]
    ) -> 'SCHISMRunConfiguration':
        required = {configuration_class: None for configuration_class in cls.REQUIRED}
        supplementary = {
            configuration_class: None for configuration_class in cls.SUPPLEMENTARY
        }

        for configuration in configurations:
            for configuration_class in required:
                if isinstance(configuration, configuration_class):
                    if required[configuration_class] is None:
                        required[configuration_class] = configuration
                        break
                    else:
                        raise ValueError(
                            f'multiple configurations given for "{configuration_class.__name__}"'
                        )
            for configuration_class in supplementary:
                if isinstance(configuration, configuration_class):
                    supplementary[configuration_class] = configuration

        instance = RunConfiguration(required.values())
        instance.__class__ = cls

        instance['modeldriver'] = required[ModelDriverJSON]
        instance['slurm'] = required[SlurmJSON]
        instance['schism'] = required[SCHISMJSON]

        forcings = [
            configuration
            for configuration in supplementary.values()
            if isinstance(configuration, ForcingJSON)
        ]
        for forcing in forcings:
            instance.add_forcing(forcing)

        return instance

    @classmethod
    def read_directory(
        cls, directory: PathLike, required: List[type] = None, supplementary: List[type] = None
    ) -> 'SCHISMRunConfiguration':
        if not isinstance(directory, Path):
            directory = Path(directory)
        if directory.is_file():
            directory = directory.parent
        if required is None:
            required = set()
        required.update(SCHISMRunConfiguration.REQUIRED)
        if supplementary is None:
            supplementary = set()
        supplementary.update(SCHISMRunConfiguration.SUPPLEMENTARY)

        return super().read_directory(directory, required, supplementary)


# class NEMSSCHISMRunConfiguration(SCHISMRunConfiguration):
#     """
#     run configuration coupling SCHISM with other models / forcings using NUOPC NEMS
#     """
#
#     REQUIRED = {
#         ModelDriverJSON,
#         NEMSJSON,
#         SlurmJSON,
#         SCHISMJSON,
#     }
#
#     def __init__(
#         self,
#         mesh_directory: PathLike,
#         modeled_start_time: datetime,
#         modeled_end_time: datetime,
#         modeled_timestep: timedelta,
#         nems_interval: timedelta,
#         nems_connections: List[str],
#         nems_mediations: List[str],
#         nems_sequence: List[str],
#         tidal_spinup_duration: timedelta = None,
#         platform: Platform = None,
#         perturbations: Dict[str, Dict[str, Any]] = None,
#         forcings: List[ForcingJSON] = None,
#         schism_processors: int = None,
#         slurm_job_duration: timedelta = None,
#         slurm_partition: str = None,
#         slurm_email_address: str = None,
#         nems_executable: PathLike = None,
#         schism_hotstart_combiner: PathLike = None,
#         schism_schout_combiner: PathLike = None,
#         # schism_use_old_io: bool = False,
#         adcprep_executable: PathLike = None,
#         aswip_executable: PathLike = None,
#         source_filename: PathLike = None,
#     ):
#         """
#         :param mesh_directory: path to input mesh directory (containing ``fort.13``, ``fort.14``)
#         :param modeled_start_time: start time within the modeled system
#         :param modeled_end_time: end time within the modeled system
#         :param modeled_timestep: time interval within the modeled system
#         :param nems_interval: modeled time interval of main NEMS loop
#         :param nems_connections: list of NEMS connections as strings (i.e. ``ATM -> OCN``)
#         :param nems_mediations: list of NEMS mediations, including functions
#         :param nems_sequence: list of NEMS entries in sequence order
#         :param schism_processors: numbers of processors to assign for SCHISM
#         :param platform: HPC platform for which to configure
#         :param tidal_spinup_duration: spinup time for SCHISM tidal coldstart
#         :param perturbations: dictionary of runs encompassing run names to parameter values
#         :param forcings: list of forcing configurations to connect to SCHISM
#         :param slurm_job_duration: wall clock time of job
#         :param slurm_partition: Slurm partition
#         :param slurm_email_address: email address to send Slurm notifications
#         :param nems_executable: filename of compiled ``schism``
#         :param source_filename: path to module file to ``source``
#         """
#
#         # TODO: Implement nems coupling for SCHISM
#
#         self.__nems = None
#
#         super().__init__(
#             mesh_directory=mesh_directory,
#             modeled_start_time=modeled_start_time,
#             modeled_end_time=modeled_end_time,
#             modeled_timestep=modeled_timestep,
#             tidal_spinup_duration=tidal_spinup_duration,
#             platform=platform,
#             perturbations=perturbations,
#             forcings=None,
#             schism_processors=schism_processors,
#             slurm_job_duration=slurm_job_duration,
#             slurm_partition=slurm_partition,
#             slurm_email_address=slurm_email_address,
#             schism_executable=nems_executable,
#             source_filename=source_filename,
#         )
#
#         nems = NEMSJSON(
#             executable_path=nems_executable,
#             modeled_start_time=modeled_start_time,
#             modeled_end_time=modeled_end_time,
#             interval=nems_interval,
#             connections=nems_connections,
#             mediations=nems_mediations,
#             sequence=nems_sequence,
#         )
#
#         self[nems.name.lower()] = nems
#
#         for forcing in forcings:
#             self.add_forcing(forcing)
#
#         self['slurm'].tasks = self.nemspy_modeling_system.processors
#
#     @property
#     def nemspy_entries(self) -> List[ModelEntry]:
#         return [
#             configuration.nemspy_entry
#             for configuration in self.configurations
#             if isinstance(configuration, NEMSCapJSON)
#         ]
#
#     @property
#     def nemspy_modeling_system(self) -> ModelingSystem:
#         return self['nems'].to_nemspy(self.nemspy_entries)
#
#     def add_forcing(self, forcing: ForcingJSON):
#         if forcing not in self:
#             forcing = self[self.add(forcing)]
#             self['schism'].add_forcing(forcing)
#
#     @classmethod
#     def from_configurations(
#         cls, configurations: List[ConfigurationJSON]
#     ) -> 'NEMSSCHISMRunConfiguration':
#         instance = SCHISMRunConfiguration.from_configurations(configurations)
#         instance.__class__ = cls
#
#         nems = None
#         for configuration in configurations:
#             if isinstance(configuration, NEMSJSON):
#                 if nems is None:
#                     nems = configuration
#                     break
#                 else:
#                     raise ValueError(
#                         f'multiple configurations given for "{NEMSJSON.__name__}"'
#                     )
#         instance['nems'] = nems
#
#         return instance
#
#     @classmethod
#     def read_directory(
#         cls, directory: PathLike, required: List[type] = None, supplementary: List[type] = None
#     ) -> 'NEMSSCHISMRunConfiguration':
#         if not isinstance(directory, Path):
#             directory = Path(directory)
#         if directory.is_file():
#             directory = directory.parent
#         if required is None:
#             required = set()
#         required.update(NEMSSCHISMRunConfiguration.REQUIRED)
#         if supplementary is None:
#             supplementary = set()
#         supplementary.update(NEMSSCHISMRunConfiguration.SUPPLEMENTARY)
#
#         return super().read_directory(directory, required, supplementary)



from argparse import ArgumentParser
from collections import Collection
import os
from pathlib import Path
from typing import Any, Dict, List

from coupledmodeldriver.client.check_completion import (
    check_completion,
    is_model_directory,
    MODELS,
)
from coupledmodeldriver.configure import ModelJSON


def parse_missing_jobs_arguments() -> Dict[str, Any]:
    argument_parser = ArgumentParser()
    argument_parser.add_argument(
        'directory',
        nargs='*',
        default=Path.cwd(),
        help='directory containing model run configuration',
    )
    argument_parser.add_argument('--model', help='model that is running, one of: `ADCIRC`')
    argument_parser.add_argument('--dependency', help='job ID of dependency')
    argument_parser.add_argument(
        '--submit', action='store_true', help='whether to submit unqueued runs'
    )

    arguments = argument_parser.parse_args()

    model = arguments.model
    if model is not None:
        model = MODELS[model.lower()]

    return {
        'directories': arguments.directory,
        'model': model,
        'submit': arguments.submit,
        'dependency': arguments.dependency,
    }


def get_run_directories(directories: List[os.PathLike], model: ModelJSON = None) -> List[Path]:
    if not isinstance(directories, Collection):
        directories = [directories]
    run_directories = []
    for directory in directories:
        if not isinstance(directory, Path):
            directory = Path(directory)
        if is_model_directory(directory, model=model):
            run_directories.append(directory)
        else:
            for subdirectory in directory.iterdir():
                if subdirectory.is_dir():
                    run_directories.extend(get_run_directories(subdirectory, model=model))

    return run_directories


def get_unqueued_runs(
    directories: List[os.PathLike], model: ModelJSON = None, **kwargs
) -> Dict[str, Path]:
    """
    get runs in the local configuration that have not been submitted / queued to the job manager

    :param directories: directory containing model run configuration
    :param model: model that is running, one of: ``ADCIRC``
    :return: mapping of unqueued run names to their directory paths
    """

    runs = {
        directory.name: directory
        for directory in get_run_directories(directories, model=model)
    }
    jobs = dict(
        line.split()[:2]
        for line in os.popen('squeue -u $USER -o "%j %Z %i" --sort i').read().splitlines()[1:]
    )

    unqueued_runs = {}
    for run_name, run_directory in runs.items():
        if run_name not in jobs:
            unqueued_runs[run_name] = run_directory

    completion_status = check_completion(unqueued_runs.values(), model=model, verbose=False)

    unqueued_runs = {
        run_name: run_directory
        for run_name, run_directory in unqueued_runs.items()
        if 'not_started' in completion_status[run_directory.name]
    }

    return unqueued_runs


def main():
    """
    submit unqueued runs in the current configuration
    """

    arguments = parse_missing_jobs_arguments()
    directories = arguments['directories']
    model = arguments['model']
    submit = arguments['submit']
    dependency = arguments['dependency']

    unqueued_runs = get_unqueued_runs(directories, model=model)
    starting_directory = Path.cwd()
    if submit:
        unqueued_run_names = []
        for run_name, run_directory in unqueued_runs.items():
            dependencies = "afterok:$(sbatch setup.job | awk '{print $NF}')"
            if dependency is not None:
                dependencies = f'{dependencies}:{dependency}'

            os.chdir(run_directory)
            os.system(f'sbatch --dependency={dependencies} adcirc.job')
            os.chdir(starting_directory)

            unqueued_run_names.append(run_name)
    else:
        unqueued_run_names = list(unqueued_runs)

    for run_name in sorted(unqueued_run_names):
        print(run_name)


if __name__ == '__main__':
    main()



from datetime import datetime, timedelta
from os import PathLike
from pathlib import Path
from typing import Any, List, Union, Dict

from nemspy.model import SCHISMEntry
from pyschism.mesh import Hgrid
from pyschism.server import SlurmConfig
from pyschism.driver import ModelConfig, ModelDriver
from pyschism.forcing.bctides.iettype import TidalElevation
from pyschism.forcing.bctides.ifltype import TidalVelocity
from pyschism.forcing.bctides.tides import Tides
from pyschism.forcing.base import ModelForcing
from pyschism.stations import Stations
from pyschism.forcing.nws.base import NWS
from pyschism.forcing import NWM

from coupledmodeldriver.configure import Model, ModelJSON, SlurmJSON
from coupledmodeldriver.configure.base import AttributeJSON, NEMSCapJSON
from coupledmodeldriver.configure.configure import from_user_input
from coupledmodeldriver.configure.forcings.base import ForcingJSON
from coupledmodeldriver.utilities import LOGGER

PYSCHISM_ATTRIBUTES = [
    'ipre',
    'ibc',
    'ibtp',
    'rnday',
    'dt',
    'nspool',
    'ihfskip',
    'ioffline_partition',
    'ipre2',
    'itransport_only',
    'iloadtide',
    'loadtide_coef',
    'start_year',
    'start_month',
    'start_day',
    'start_hour',
    'utc_start',
    'ics',
    'ihot',
    'ieos_type',
    'ieos_pres',
    'eos_a',
    'eos_b',
    'dramp',
    'drampbc',
    'iupwind_mom',
    'indvel',
    'ihorcon',
    'hvis_coef0',
    'ishapiro',
    'niter_shap',
    'shapiro0',
    'thetai',
    'icou_elfe_wwm',
    'nstep_wwm',
    'iwbl',
    'hmin_radstress',
    'drampwafo',
    'turbinj',
    'turbinjds',
    'alphaw',
    'fwvor_advxy_stokes',
    'fwvor_advz_stokes',
    'fwvor_gradpress',
    'fwvor_breaking',
    'fwvor_streaming',
    'cur_wwm',
    'wafo_obcramp',
    'imm',
    'ibdef',
    'slam0',
    'sfea0',
    'iunder_deep',
    'h1_bcc',
    'h2_bcc',
    'hw_depth',
    'hw_ratio',
    'ihydraulics',
    'if_source',
    'dramp_ss',
    'meth_sink',
    'ihdif',
    'nchi',
    'dzb_min',
    'hmin_man',
    'ncor',
    'rlatitude',
    'coricoef',
    'ic_elev',
    'nramp_elev',
    'inv_atm_bnd',
    'prmsl_ref',
    'gen_wsett',
    'ibcc_mean',
    'rmaxvel',
    'velmin_btrack',
    'ihhat',
    'inunfl',
    'h0',
    'shorewafo',
    'moitn0',
    'mxitn0',
    'rtol0',
    'nadv',
    'dtb_max',
    'dtb_min',
    'inter_mom',
    'kr_co',
    'itr_met',
    'h_tvd',
    'eps1_tvd_imp',
    'eps2_tvd_imp',
    'ielm_transport',
    'max_subcyc',
    'ip_weno',
    'nquad',
    'ntd_weno',
    'epsilon1',
    'epsilon2',
    'i_prtnftl_weno',
    'epsilon3',
    'ielad_weno',
    'small_elad',
    'nws',
    'wtiminc',
    'drampwind',
    'iwindoff',
    'iwind_form',
    'impose_net_flux',
    'ihconsv',
    'isconsv',
    'i_hmin_airsea_ex',
    'hmin_airsea_ex',
    'i_hmin_salt_ex',
    'hmin_salt_ex',
    'itur',
    'dfv0',
    'dfh0',
    'mid',
    'stab',
    'xlsc0',
    'inu_elev',
    'inu_uv',
    'vnh1',
    'vnf1',
    'vnh2',
    'vnf2',
    'step_nu_tr',
    'h_bcc1',
    's1_mxnbt',
    's2_mxnbt',
    'iharind',
    'iflux',
    'izonal5',
    'ibtrack_test',
    'irouse_test',
    'flag_fib',
    'slr_rate',
    'isav',
    'nstep_ice',
    'level_age',
    'rearth_pole',
    'rearth_eq',
    'shw',
    'rho0',
    'h_massconsv',
    'rinflation_icm',
    'nhot',
    'nhot_write',
    'iout_sta',
    'nspool_sta',
]
OUTPUT_INTERVAL_DEFAULTS = {
    'surface_output_interval': timedelta(hours=1),
    'stations_output_interval': timedelta(minutes=6),
}


class SCHISMJSON(ModelJSON, NEMSCapJSON, AttributeJSON):
    """
    SCHISM configuration in ``configure_schism.json``

    stores a number of SCHISM parameters and optionally NEMS parameters
    """

    name = 'SCHISM'
    default_filename = f'configure_schism.json'
    default_processors = 11
    default_attributes = PYSCHISM_ATTRIBUTES

    field_types = {
        'schism_executable_path': Path,
        'schism_hotstart_combiner_path': Path,
        'schism_schout_combiner_path': Path,
        'modeled_start_time': datetime,
        'modeled_end_time': datetime,
        'modeled_timestep': timedelta,
        'fgrid_path': Path,
        'hgrid_path': Path,
        'schism_use_old_io': bool,
        'tidal_spinup_duration': timedelta,
        'tidal_spinup_timestep': timedelta,
        'source_filename': Path,
        'use_original_mesh': bool,
        'output_surface': bool,
        'surface_output_interval': timedelta,
        'output_stations': bool,
        'stations_file_path': Path,
        'stations_output_interval': timedelta,
        'output_spinup': bool,
        'output_elevations': bool,
        #        'output_velocities': bool,
        #        'output_concentrations': bool,
    }

    def __init__(
        self,
        schism_executable_path: PathLike,
        schism_hotstart_combiner_path: PathLike,
        modeled_start_time: datetime,
        modeled_end_time: datetime,
        modeled_timestep: timedelta,
        fgrid_path: PathLike,
        hgrid_path: PathLike,
        schism_use_old_io: bool = False,
        schism_schout_combiner_path: PathLike = None,
        tidal_spinup_duration: timedelta = None,
        tidal_spinup_timestep: timedelta = None,
        forcings: List[ModelForcing] = None,
        source_filename: PathLike = None,
        slurm_configuration: SlurmJSON = None,
        use_original_mesh: bool = False,
        output_surface: bool = True,
        surface_output_interval: timedelta = None,
        output_stations: bool = False,
        stations_output_interval=None,
        stations_file_path: PathLike = None,
        output_spinup: bool = True,
        output_elevations: bool = True,
        output_velocities: bool = True,
        #        output_concentrations: bool = False,
        #        output_meteorological_factors: bool = False,
        processors: int = 11,
        nems_parameters: Dict[str, str] = None,
        attributes: Dict[str, Any] = None,
        **kwargs,
    ):
        """
        :param schism_executable_path: file path to ``pschism_TVD-VL`` or ``NEMS.x``
        :param schism_hotstart_combiner_path: file path to ``combine_hotstart7``
        :param schism_schout_combiner_path: file path to ``combine_output11``
        :param modeled_start_time: start time in model run
        :param modeled_end_time: edn time in model run
        :param modeled_timestep: time interval between model steps
        :param fgrid_path: file path to bottom friction file
        :param hgrid_path: file path to ``hgrid.gr3``
        :param schism_use_old_io: flag to use old or new schism IO
        :param tidal_spinup_duration: tidal spinup duration for SCHISM coldstart
        :param tidal_spinup_timestep: tidal spinup modeled time interval for SCHISM coldstart
        :param forcings: list of ModelForcing objects to apply to the mesh
        :param source_filename: path to modulefile to ``source``
        :param slurm_configuration: Slurm configuration object
        :param use_original_mesh: whether to symlink / copy original mesh instead of rewriting with ``pyschism``
        :param output_surface: write surface (entire mesh) to NetCDF
        :param surface_output_interval: frequency at which output is written to file
        :param output_stations: write stations outputs timehistory (only applicable if stations file exists)
        :param stations_output_interval: frequency at which stations output is written to file
        :param stations_file_path: file path to stations file
        :param output_spinup: write spinup to NetCDF
        :param output_elevations: write elevations to NetCDF
        :param output_velocities: write velocities to NetCDF
        :param processors: number of processors to use
        :param nems_parameters: parameters to give to NEMS cap
        :param attributes: attributes to set in ``pyschism.Param`` object
        """

        self.__mesh = None
        self.__base_mesh = None

        if tidal_spinup_timestep is None:
            tidal_spinup_timestep = modeled_timestep

        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(SCHISMJSON.field_types)

        ModelJSON.__init__(self, model=Model.SCHISM, **kwargs)
        NEMSCapJSON.__init__(
            self, processors=processors, nems_parameters=nems_parameters, **kwargs
        )
        AttributeJSON.__init__(self, attributes=attributes, **kwargs)

        self['schism_executable_path'] = schism_executable_path
        self['schism_hotstart_combiner_path'] = schism_hotstart_combiner_path
        self['schism_schout_combiner_path'] = schism_schout_combiner_path
        self['modeled_start_time'] = modeled_start_time
        self['modeled_end_time'] = modeled_end_time
        self['modeled_timestep'] = modeled_timestep
        self['fgrid_path'] = fgrid_path
        self['hgrid_path'] = hgrid_path
        self['schism_use_old_io'] = schism_use_old_io
        self['tidal_spinup_duration'] = tidal_spinup_duration
        self['tidal_spinup_timestep'] = tidal_spinup_timestep
        self['source_filename'] = source_filename
        self['use_original_mesh'] = use_original_mesh

        self['output_surface'] = output_surface
        self['surface_output_interval'] = surface_output_interval
        self['output_stations'] = output_stations
        self['stations_output_interval'] = stations_output_interval
        self['stations_file_path'] = stations_file_path
        self['output_spinup'] = output_spinup
        self['output_elevations'] = output_elevations
        self['output_velocities'] = output_velocities
        #        self['output_concentrations'] = output_concentrations
        #        self['output_meteorological_factors'] = output_meteorological_factors

        self.__forcings = []
        self.forcings = forcings
        self.slurm_configuration = slurm_configuration

        for output_interval_entry, default_interval in OUTPUT_INTERVAL_DEFAULTS.items():
            if self[output_interval_entry] is None:
                LOGGER.debug(f'setting `{output_interval_entry}` to {default_interval}')
                self[output_interval_entry] = default_interval

    @property
    def forcings(self) -> List[ForcingJSON]:
        return list(self.__forcings)

    @forcings.setter
    def forcings(self, forcings: List[ForcingJSON]):
        if forcings is None:
            forcings = []
        for forcing in forcings:
            self.add_forcing(forcing)

    def add_forcing(self, forcing: ForcingJSON):
        if not isinstance(forcing, ForcingJSON):
            forcing = from_user_input(forcing)
        pyschism_forcing = forcing.pyschism_forcing

        existing_forcings = [
            existing_forcing.__class__.__name__ for existing_forcing in self.pyschism_forcings
        ]
        if pyschism_forcing.__class__.__name__ in existing_forcings:
            existing_index = existing_forcings.index(pyschism_forcing.__class__.__name__)
        else:
            existing_index = -1
        if existing_index > -1:
            self.__forcings[existing_index] = forcing
        else:
            self.__forcings.append(forcing)

    @property
    def slurm_configuration(self) -> List[SlurmJSON]:
        return self.__slurm_configuration

    @slurm_configuration.setter
    def slurm_configuration(self, slurm_configuration: SlurmJSON):
        if isinstance(slurm_configuration, SlurmConfig):
            SlurmJSON.from_pyschism(slurm_configuration)
        self.__slurm_configuration = slurm_configuration

    @property
    def pyschism_forcings(self) -> List[ModelForcing]:
        return [forcing.pyschism_forcing for forcing in self.forcings]

    @property
    def pyschism_mesh(self) -> Hgrid:
        if self.__mesh is None:
            # TODO: PySCHISM copy() for hgrid loses boundary information!
            #            mesh = self.base_mesh.copy()
            mesh = self.base_mesh

            # TODO: reconstruct mesh from base?
            self.__mesh = mesh

        return self.__mesh

    @pyschism_mesh.setter
    def pyschism_mesh(self, pyschism_mesh: Union[Hgrid, PathLike]):
        self.__mesh = pyschism_mesh

    @property
    def base_mesh(self) -> Hgrid:
        if self.__base_mesh is None:
            if self.__mesh is not None:
                self.__base_mesh = self.__mesh
            else:
                self.__base_mesh = self['hgrid_path']
        if not isinstance(self.__base_mesh, Hgrid):
            LOGGER.info(f'opening mesh "{self.__base_mesh}"')
            self.__base_mesh = Hgrid.open(self.__base_mesh, crs=4326)
        # TODO: deconstruct mesh into a base mesh that can be pickled
        return self.__base_mesh

    @base_mesh.setter
    def base_mesh(self, base_mesh: Union[Hgrid, PathLike]):
        self.__base_mesh = base_mesh

    @property
    def pyschism_driver(self) -> ModelDriver:

        tidal_elevation = None
        tidal_velocity = None
        meteo = None
        hydrology = None
        for pyschism_forcing in self.pyschism_forcings:
            if isinstance(pyschism_forcing, Tides):

                # NOTE: Create tidal BC and then replace the tidal database
                tidal_elevation = TidalElevation()
                tidal_elevation.tides = pyschism_forcing
                tidal_velocity = TidalVelocity()
                tidal_velocity.tides = pyschism_forcing

            elif isinstance(pyschism_forcing, NWS):
                meteo = pyschism_forcing

            elif isinstance(pyschism_forcing, NWM):
                hydrology = pyschism_forcing

        config = ModelConfig(
            hgrid=self.pyschism_mesh,
            vgrid=None,
            fgrid=None,  # pyschism writes linear with depth for 2D
            iettype=tidal_elevation,
            ifltype=tidal_velocity,
            nws=meteo,
            source_sink=hydrology,
        )

        # TODO: What about other variable outputs?

        stations = None
        if self['stations_file_path'] is not None:
            stations_output_interval = self['stations_output_interval']
            if stations_output_interval is None:
                stations_output_interval = self['surface_output_interval']

            stations = Stations.from_file(
                file=self['stations_file_path'],
                nspool_sta=stations_output_interval,
                crs=4326,
                elev=self['output_elevations'],
                u=self['output_velocities'],
                v=self['output_velocities'],
            )

        spinup_duration = self['tidal_spinup_duration']
        if spinup_duration is None:
            spinup_duration = timedelta(0)

        # Note we always create coldstart driver, if we want to write
        # hotstart, we just remove ramp attribtues later
        driver = config.coldstart(
            start_date=self['modeled_start_time'],
            end_date=self['modeled_end_time'],
            timestep=self['modeled_timestep'],
            dramp=spinup_duration,
            drampbc=spinup_duration,
            dramp_ss=spinup_duration,
            drampwind=spinup_duration,
            nspool=self['surface_output_interval'],
            elev=self['output_elevations'],
            dahv=self['output_velocities'],
            stations=stations,
            ihfskip=None,  # pyschism sets None to "rnday" value
            nhot_write=None,  # pyschism sets None to "ihfskip" value
        )

        # Add manual overrides to parameters in param.nml
        for name, value in self['attributes'].items():
            if value is not None:
                try:
                    for nml_name in ['core', 'opt', 'schout']:
                        nml_attr = getattr(driver.param, nml_name)
                        if name not in dir(nml_attr):
                            continue

                        setattr(nml_attr, name, value)
                except:
                    LOGGER.warning(
                        f'could not set `{driver.__class__.__name__}` attribute `{name}` to `{value}`'
                    )

        return driver

    @property
    def nemspy_entry(self) -> SCHISMEntry:
        return SCHISMEntry(processors=self['processors'], **self['nems_parameters'])

    def __setitem__(self, key: str, value: Any):
        super().__setitem__(key, value)
        self.__mesh = None

    def __copy__(self) -> 'SCHISMJSON':
        instance = super().__copy__()
        instance.forcings = self.forcings
        return instance



from argparse import ArgumentParser
from datetime import datetime, timedelta
from enum import Enum
import logging
import os
from pathlib import Path
from typing import Any, Dict, List, Mapping, Tuple

from typepigeon import convert_value
from pyschism.forcing.bctides.tides import TidalDatabase

from coupledmodeldriver import Platform
from coupledmodeldriver.configure import (
    ATMESHForcingJSON,
    BestTrackForcingJSON,
    OWIForcingJSON,
    TidalForcingJSON,
    WW3DATAForcingJSON,
    NationalWaterModelFocringJSON,
)
from coupledmodeldriver.configure.base import NEMSCapJSON
from coupledmodeldriver.configure.forcings.base import (
    FileForcingJSON,
    ForcingJSON,
    TimestepForcingJSON,
    WaveForcingJSON,
    WindForcingJSON,
)
from coupledmodeldriver.generate import (
    SCHISMRunConfiguration,
    # NEMSSCHISMRunConfiguration,
)
from coupledmodeldriver.generate.schism.script import SchismEnsembleGenerationJob
from coupledmodeldriver.utilities import get_logger


class ForcingConfigurations(Enum):
    tidal = TidalForcingJSON
    besttrack = BestTrackForcingJSON
    nwm = NationalWaterModelFocringJSON


FORCING_NAMES = list(entry.name for entry in ForcingConfigurations)
DEFAULT_TIDAL_SOURCE = TidalDatabase.TPXO
DEFAULT_TIDAL_CONSTITUENTS = 'all'


def parse_initialize_schism_arguments(
    extra_arguments: Dict[str, Tuple[type, str]] = None
) -> Dict[str, Any]:
    if extra_arguments is None:
        extra_arguments = {}
    elif not isinstance(extra_arguments, Mapping):
        extra_arguments = {extra_argument: (None, None) for extra_argument in extra_arguments}
    extra_arguments = {
        extra_argument.strip('-'): argument_info
        for extra_argument, argument_info in extra_arguments.items()
    }

    argument_parser = ArgumentParser()

    argument_parser.add_argument(
        '--platform', required=True, help='HPC platform for which to configure'
    )
    argument_parser.add_argument(
        '--mesh-directory',
        required=True,
        help='path to input mesh (`hgrid.gr3`, `manning.gr3` or `drag.gr3`)',
    )
    argument_parser.add_argument(
        '--modeled-start-time', required=True, help='start time within the modeled system'
    )
    argument_parser.add_argument(
        '--modeled-duration', required=True, help='duration within the modeled system'
    )
    argument_parser.add_argument(
        '--modeled-timestep', required=True, help='time interval within the modeled system'
    )
    argument_parser.add_argument(
        '--nems-interval', default=None, help='main loop interval of NEMS run'
    )
    argument_parser.add_argument(
        '--modulefile',
        default=None,
        help='path to modulefile to source before model execution`',
    )
    argument_parser.add_argument(
        '--forcings',
        default=None,
        help=f'comma-separated list of forcings to configure, from {FORCING_NAMES}',
    )
    argument_parser.add_argument(
        '--schism-executable',
        default='pschism-TVD_VL',
        help='filename of compiled `pschism-TVD_VL` or `NEMS.x`',
    )
    argument_parser.add_argument(
        '--schism-hotstart-combiner',
        default='combine_hotstart7',
        help='filename of compiled `combine_hotstart7` executable',
    )
    argument_parser.add_argument(
        '--schism-schout-combiner',
        default='combine_output11',
        help='filename of compiled `combine_output11` executable',
    )
    argument_parser.add_argument(
        '--schism-use-old-io',
        action='store_true',
        help='flag to indicate whether SCHISM executable uses old or new IO',
    )
    argument_parser.add_argument(
        '--schism-processors', default=11, help='numbers of processors to assign for SCHISM'
    )
    argument_parser.add_argument(
        '--job-duration', default='06:00:00', help='wall clock time for job'
    )
    argument_parser.add_argument(
        '--output-directory',
        default=Path().cwd(),
        help='directory to which to write configuration files (defaults to `.`)',
    )
    argument_parser.add_argument(
        '--skip-existing', action='store_true', help='skip existing files',
    )
    argument_parser.add_argument(
        '--absolute-paths',
        action='store_true',
        help='write paths as absolute in configuration',
    )
    argument_parser.add_argument(
        '--verbose', action='store_true', help='show more verbose log messages'
    )

    # add extra arguments with bool types and descriptions
    for extra_argument, (argument_type, argument_description) in extra_arguments.items():
        kwargs = {}
        if argument_type is bool:
            kwargs['action'] = 'store_true'
        if argument_description is not None:
            kwargs['help'] = argument_description
        argument_parser.add_argument(f'--{extra_argument}', **kwargs)

    arguments, unknown_arguments = argument_parser.parse_known_args()

    # convert extra arguments to their given type
    extra_arguments = {
        extra_argument: convert_value(
            arguments.__dict__[extra_argument.replace('-', '_')], argument_info[0]
        )
        for extra_argument, argument_info in extra_arguments.items()
    }

    platform = convert_value(arguments.platform, Platform)
    mesh_directory = convert_value(arguments.mesh_directory, Path).resolve().absolute()

    modeled_start_time = convert_value(arguments.modeled_start_time, datetime)
    modeled_duration = convert_value(arguments.modeled_duration, timedelta)
    modeled_timestep = convert_value(arguments.modeled_timestep, timedelta)
    nems_interval = convert_value(arguments.nems_interval, timedelta)

    schism_processors = convert_value(arguments.schism_processors, int)

    modulefile = convert_value(arguments.modulefile, Path)
    if modulefile is not None:
        modulefile = modulefile.resolve().absolute()

    forcings = arguments.forcings
    if forcings is not None:
        forcings = [forcing.strip() for forcing in forcings.split(',')]
    else:
        forcings = []

    schism_executable = convert_value(arguments.schism_executable, Path).resolve().absolute()
    schism_hotstart_combiner = (
        convert_value(arguments.schism_hotstart_combiner, Path).resolve().absolute()
    )
    schism_schout_combiner = (
        convert_value(arguments.schism_schout_combiner, Path).resolve().absolute()
    )
    schism_use_old_io = convert_value(arguments.schism_use_old_io, bool)

    job_duration = convert_value(arguments.job_duration, timedelta)
    output_directory = convert_value(arguments.output_directory, Path).resolve().absolute()

    absolute_paths = arguments.absolute_paths
    overwrite = not arguments.skip_existing
    verbose = arguments.verbose

    arguments = {}
    unrecognized_arguments = []
    for index in range(len(unknown_arguments)):
        argument = unknown_arguments[index]
        value = None
        if argument.startswith('-'):
            parsed_argument = argument.strip('-').strip()
            if len(unknown_arguments) > index + 1 and not unknown_arguments[
                index + 1
            ].startswith('-'):
                value = unknown_arguments[index + 1].strip()
            forcing = parsed_argument.split('-')[0]
            if forcing not in forcings:
                if forcing.lower() in FORCING_NAMES:
                    forcings.append(forcing.lower())
                else:
                    unrecognized_arguments.append(argument)
            arguments[parsed_argument] = value

    if len(unrecognized_arguments) > 0:
        argument_parser.error(f'unrecognized arguments: {" ".join(unrecognized_arguments)}')

    unknown_arguments = arguments
    del arguments

    tidal_spinup_duration = None

    # initialize `pyschism` forcing objects
    forcing_configurations = []
    for provided_name in forcings:
        if provided_name.lower() in FORCING_NAMES:
            forcing_configuration_class = ForcingConfigurations[provided_name.lower()].value
            kwargs = {}
            if issubclass(forcing_configuration_class, TidalForcingJSON):
                tidal_spinup_duration = get_argument(
                    argument=f'tidal-spinup-duration',
                    arguments=unknown_arguments,
                    required=True,
                    message=f'enter tidal spinup duration (`HH:MM:SS`): ',
                )
                tidal_spinup_duration = convert_value(tidal_spinup_duration, timedelta)
                tidal_source_options = '/'.join(
                    tidal_source.name.lower()
                    if tidal_source != DEFAULT_TIDAL_SOURCE
                    else tidal_source.name.upper()
                    for tidal_source in TidalDatabase
                )
                tidal_source = get_argument(
                    argument=f'tidal-source',
                    arguments=unknown_arguments,
                    required=True,
                    message=f'enter tidal forcing source ({tidal_source_options}): ',
                )
                if tidal_source is not None:
                    tidal_source = convert_value(tidal_source, TidalDatabase)
                else:
                    tidal_source = DEFAULT_TIDAL_SOURCE
                tidal_constituents = get_argument(
                    argument='tidal-constituents', arguments=unknown_arguments,
                )
                if tidal_constituents is not None:
                    tidal_constituents = [
                        entry.strip() for entry in tidal_constituents.split(',')
                    ]
                else:
                    tidal_constituents = DEFAULT_TIDAL_CONSTITUENTS
                kwargs['tidal_source'] = tidal_source
                kwargs['constituents'] = tidal_constituents
            if issubclass(forcing_configuration_class, BestTrackForcingJSON):
                nhc_code = get_argument(
                    argument='besttrack-nhc-code',
                    arguments=unknown_arguments,
                    required=True,
                    message='enter NHC code of storm: ',
                )
                kwargs['nhc_code'] = nhc_code
                best_track_start_date = get_argument(
                    argument='besttrack-start-date', arguments=unknown_arguments,
                )
                kwargs['start_date'] = best_track_start_date
                best_track_end_date = get_argument(
                    argument='besttrack-end-date', arguments=unknown_arguments,
                )
                kwargs['end_date'] = best_track_end_date

            if issubclass(forcing_configuration_class, FileForcingJSON):
                forcing_path = get_argument(
                    argument=f'{provided_name}-path', arguments=unknown_arguments,
                )
                kwargs['resource'] = forcing_path
            if issubclass(forcing_configuration_class, NEMSCapJSON):
                nems_cap_processors = get_argument(
                    argument=f'{provided_name}-processors', arguments=unknown_arguments,
                )
                kwargs['processors'] = nems_cap_processors
                nems_cap_parameters = get_argument(
                    argument=f'{provided_name}-nems-parameters', arguments=unknown_arguments,
                )
                kwargs['nems_parameters'] = nems_cap_parameters
            if issubclass(forcing_configuration_class, TimestepForcingJSON):
                forcing_interval = get_argument(
                    argument=f'{provided_name}-interval', arguments=unknown_arguments,
                )
                if forcing_interval is None:
                    forcing_interval = nems_interval
                kwargs['interval'] = forcing_interval
            if issubclass(forcing_configuration_class, WindForcingJSON):
                forcing_nws = get_argument(
                    argument=f'{provided_name}-nws', arguments=unknown_arguments,
                )
                kwargs['nws'] = forcing_nws
            if issubclass(forcing_configuration_class, WaveForcingJSON):
                forcing_nrs = get_argument(
                    argument=f'{provided_name}-nrs', arguments=unknown_arguments,
                )
                kwargs['nrs'] = forcing_nrs

            forcing_configurations.append(forcing_configuration_class(**kwargs))
        else:
            raise NotImplementedError(
                f'unrecognized forcing "{provided_name}"; must be from {FORCING_NAMES}'
            )

    return {
        'platform': platform,
        'mesh_directory': mesh_directory,
        'modeled_start_time': modeled_start_time,
        'modeled_duration': modeled_duration,
        'modeled_timestep': modeled_timestep,
        'nems_interval': nems_interval,
        'tidal_spinup_duration': tidal_spinup_duration,
        'perturbations': None,
        'modulefile': modulefile,
        'forcings': forcing_configurations,
        'schism_executable': schism_executable,
        'schism_hotstart_combiner': schism_hotstart_combiner,
        'schism_schout_combiner': schism_schout_combiner,
        'schism_use_old_io': schism_use_old_io,
        'schism_processors': schism_processors,
        'job_duration': job_duration,
        'output_directory': output_directory,
        'absolute_paths': absolute_paths,
        'overwrite': overwrite,
        'verbose': verbose,
        **{extra_argument: value for extra_argument, value in extra_arguments.items()},
    }


def initialize_schism(
    platform: Platform,
    mesh_directory: os.PathLike,
    modeled_start_time: datetime,
    modeled_duration: timedelta,
    modeled_timestep: timedelta,
    tidal_spinup_duration: timedelta = None,
    perturbations: Dict[str, Dict[str, Any]] = None,
    nems_interval: timedelta = None,
    nems_connections: List[str] = None,
    nems_mediations: List[str] = None,
    nems_sequence: List[str] = None,
    modulefile: os.PathLike = None,
    forcings: List[ForcingJSON] = None,
    schism_executable: os.PathLike = None,
    schism_hotstart_combiner: os.PathLike = None,
    schism_schout_combiner: os.PathLike = None,
    schism_use_old_io: bool = False,
    schism_processors: int = None,
    job_duration: timedelta = None,
    output_directory: os.PathLike = None,
    absolute_paths: bool = True,
    overwrite: bool = None,
    verbose: bool = False,
):
    """
    creates a set of JSON configuration files according to the given parameters

    :param platform: HPC platform for which to configure
    :param mesh_directory: path to input mesh (`fort.13`, `fort.14`)
    :param modeled_start_time: start time within the modeled system
    :param modeled_duration: duration within the modeled system
    :param modeled_timestep: time interval within the modeled system
    :param tidal_spinup_duration: duration of tidal spinup in model time
    :param perturbations: mapping of perturbation names to changed values
    :param nems_interval: modeled time interval of NEMS main loop
    :param nems_connections: list of NEMS connections as strings (i.e. ``ATM -> OCN``)
    :param nems_mediations: list of NEMS mediations, including functions
    :param nems_sequence: list of NEMS entries in sequence order
    :param modulefile: path to modulefile to source before model execution
    :param forcings: list of forcings to configure, from ['tidal', 'atmesh', 'besttrack', 'owi', 'ww3data']
    :param schism_executable: filename of compiled ``pschism-TVD_VL` `or ``NEMS.x``
    :param schism_hotstart_combiner: filename of compiled hotstart combiner
    :param schism_schout_combiner: filename of compiled SCHISM old output combiner
    :param schism_use_old_io: flag to indicate if the compiled SCHISM uses old IO
    :param schism_processors: numbers of processors to assign for SCHISM
    :param job_duration: wall clock time for job
    :param output_directory: directory to which to write configuration files (defaults to ``.``)
    :param absolute_paths: whether to use absolute paths in configuration
    :param overwrite: whether to overwrite existing files
    :param verbose: whether to log extra debugging messages
    """

    logger = get_logger(
        'initialize_schism', console_level=logging.DEBUG if verbose else logging.INFO
    )

    if nems_interval is not None:
        raise ValueError('NEMS run setup is not supported for SCHISM yet!')

        # configuration = NEMSSCHISMRunConfiguration(
        #     mesh_directory=mesh_directory,
        #     modeled_start_time=modeled_start_time,
        #     modeled_end_time=modeled_start_time + modeled_duration,
        #     modeled_timestep=modeled_timestep,
        #     nems_interval=nems_interval,
        #     nems_connections=nems_connections,
        #     nems_mediations=nems_mediations,
        #     nems_sequence=nems_sequence,
        #     tidal_spinup_duration=tidal_spinup_duration,
        #     platform=platform,
        #     perturbations=perturbations,
        #     forcings=forcings,
        #     schism_processors=schism_processors,
        #     slurm_partition=None,
        #     slurm_job_duration=job_duration,
        #     slurm_email_address=None,
        #     nems_executable=schism_executable,
        #     schism_hotstart_combiner=schism_hotstart_combiner,
        #     schism_schout_combiner=schism_schout_combiner,
        #     schism_use_old_io=schism_use_old_io,
        #     source_filename=modulefile,
        # )
    else:
        configuration = SCHISMRunConfiguration(
            mesh_directory=mesh_directory,
            modeled_start_time=modeled_start_time,
            modeled_end_time=modeled_start_time + modeled_duration,
            modeled_timestep=modeled_timestep,
            tidal_spinup_duration=tidal_spinup_duration,
            platform=platform,
            perturbations=perturbations,
            forcings=forcings,
            schism_processors=schism_processors,
            slurm_partition=None,
            slurm_job_duration=job_duration,
            slurm_email_address=None,
            schism_executable=schism_executable,
            schism_hotstart_combiner=schism_hotstart_combiner,
            schism_schout_combiner=schism_schout_combiner,
            schism_use_old_io=schism_use_old_io,
            source_filename=modulefile,
        )

    if not absolute_paths:
        output_directory = Path(os.path.relpath(output_directory, Path.cwd()))
        configuration.relative_to(output_directory, inplace=True)

    components = [
        configuration_json.name for configuration_json in configuration.configurations
    ]

    logger.info(f'writing {"+".join(components)} configuration to "{output_directory}"')

    for configuration_json in configuration.configurations:
        logger.debug(repr(configuration_json))

    configuration.write_directory(
        directory=output_directory, absolute=absolute_paths, overwrite=overwrite,
    )

    if platform == Platform.HERA:
        partition = 'bigmem'
    else:
        partition = None

    generation_job_script = SchismEnsembleGenerationJob(
        platform=platform, parallel=True, slurm_partition=partition
    )
    generation_job_script.write(filename=output_directory / 'generate.job', overwrite=True)


def get_argument(
    argument: str,
    arguments: Dict[str, str] = None,
    required: bool = False,
    message: str = None,
) -> str:
    if message is None:
        message = f'enter value for "{argument}": '

    if argument in arguments:
        value = arguments[argument]
    elif required:
        value = input(message)
    else:
        value = ''

    if len(value.strip()) == 0:
        value = None

    return value


def main():
    initialize_schism(**parse_initialize_schism_arguments())


if __name__ == '__main__':
    main()



from datetime import timedelta
from os import PathLike
from pathlib import Path
from typing import List

from coupledmodeldriver.platforms import Platform
from coupledmodeldriver.script import (
    JobScript,
    EnsembleGenerationJob,
    EnsembleRunScript,
    EnsembleCleanupScript,
)


class AdcircJob(JobScript):
    """
    abstraction of a job script for running ADCIRC
    """

    def __init__(
        self,
        platform: Platform,
        commands: List[str],
        slurm_run_name: str = None,
        slurm_tasks: int = None,
        slurm_duration: timedelta = None,
        slurm_account: str = None,
        source_filename: PathLike = None,
        **kwargs,
    ):
        if slurm_run_name is None:
            slurm_run_name = 'ADCIRC_JOB'

        super().__init__(
            platform=platform,
            commands=commands,
            slurm_run_name=slurm_run_name,
            slurm_tasks=slurm_tasks,
            slurm_duration=slurm_duration,
            slurm_account=slurm_account,
            **kwargs,
        )

        if source_filename is not None:
            if not isinstance(source_filename, Path):
                source_filename = Path(source_filename)
            source_filename = source_filename.as_posix()
            if len(str(source_filename)) > 0:
                self.commands.insert(0, f'source {source_filename}')


class AdcircRunJob(AdcircJob):
    """
    job script for running ADCIRC
    """

    def __init__(
        self,
        platform: Platform,
        slurm_run_name: str = None,
        slurm_tasks: int = None,
        slurm_duration: timedelta = None,
        slurm_account: str = None,
        executable: PathLike = None,
        commands: List[str] = None,
        **kwargs,
    ):
        super().__init__(
            platform=platform,
            commands=commands,
            slurm_run_name=slurm_run_name,
            slurm_tasks=slurm_tasks,
            slurm_duration=slurm_duration,
            slurm_account=slurm_account,
            **kwargs,
        )

        if executable is None:
            executable = 'adcirc'
        else:
            if not isinstance(executable, Path):
                executable = Path(executable)
            executable = executable.as_posix()

        self.executable = executable
        self.commands.append(
            f'{self.launcher} {self.executable}'
            if self.launcher is not None
            else f'{self.executable}'
        )


class AswipCommand:
    """
    abstraction of an ``aswip`` command

    https://wiki.adcirc.org/wiki/Fort.22_file
    """

    def __init__(
        self,
        path: PathLike,
        nws: int = None,
        isotachs: int = None,
        rmax_approaches: int = None,
    ):
        self.__path = None
        self.__nws = None
        self.__isotachs = None
        self.__rmax_approaches = None

        self.path = path
        self.nws = nws
        self.isotachs = isotachs
        self.rmax_approaches = rmax_approaches

    @property
    def path(self) -> Path:
        """
        path to ``aswip`` executable
        """
        return self.__path

    @path.setter
    def path(self, path: PathLike):
        if not isinstance(path, Path):
            path = Path(path)

        self.__path = path

    @property
    def nws(self) -> int:
        """
        ``-n`` - ADCIRC NWS option
        """
        return self.__nws

    @nws.setter
    def nws(self, nws: int):
        self.__nws = nws

    @property
    def isotachs(self) -> int:
        """
        ``-m`` - methods using isotachs, one of [1, 2, 3, 4]

        1 = use the 34 isotach,
        2 = use the 64 isotach,
        3 = use the 50 isotach,
        4 = use all available isotachs (use 4 for NWS=20)
        """

        if self.__isotachs is None:
            if self.nws == 20:
                self.__isotachs = 4

        return self.__isotachs

    @isotachs.setter
    def isotachs(self, isotachs: int):
        self.__isotachs = isotachs

    @property
    def rmax_approaches(self) -> int:
        """
        ``-z`` - approaches solving for Rmax, one of [1, 2]

        1 = only rotate wind vectors afterward,
        2 = rotate wind vectors before and afterwards (use this for NWS=20)
        """

        if self.__rmax_approaches is None:
            if self.nws == 20:
                self.__rmax_approaches = 2

        return self.__rmax_approaches

    @rmax_approaches.setter
    def rmax_approaches(self, rmax_approaches: int):
        self.__rmax_approaches = rmax_approaches

    def __str__(self) -> str:
        aswip_command = f'{self.path.as_posix()} -n {self.nws}'

        if self.isotachs is not None:
            aswip_command = f'{aswip_command} -m {self.isotachs}'
        if self.isotachs is not None:
            aswip_command = f'{aswip_command} -z {self.rmax_approaches}'

        return '\n'.join(
            [f'{aswip_command}', 'mv fort.22 fort.22.original', 'mv NWS_20_fort.22 fort.22']
        )

    @classmethod
    def from_string(cls, aswip_command: str) -> 'AswipCommand':
        parts = aswip_command.split()[0]
        path = parts[0]

        if '-n' in parts:
            nws = parts[parts.index('-n') + 1]
        else:
            nws = None

        if '-m' in parts:
            isotachs = parts[parts.index('-m') + 1]
        else:
            isotachs = None

        if '-z' in parts:
            rmax_approaches = parts[parts.index('-z') + 1]
        else:
            rmax_approaches = None

        return cls(path=path, nws=nws, isotachs=isotachs, rmax_approaches=rmax_approaches)


class AdcircSetupJob(AdcircJob):
    """
    job script for performing domain decomposition with ``adcprep``
    """

    def __init__(
        self,
        platform: Platform,
        adcirc_mesh_partitions: int,
        slurm_run_name: str,
        slurm_tasks: int = None,
        slurm_duration: timedelta = None,
        slurm_account: str = None,
        adcprep_path: PathLike = None,
        aswip_command: str = None,
        commands: List[str] = None,
        **kwargs,
    ):
        if slurm_tasks is None:
            slurm_tasks = 1

        super().__init__(
            platform=platform,
            commands=commands,
            slurm_run_name=slurm_run_name,
            slurm_tasks=slurm_tasks,
            slurm_duration=slurm_duration,
            slurm_account=slurm_account,
            **kwargs,
        )

        self.adcirc_partitions = adcirc_mesh_partitions

        if adcprep_path is None:
            adcprep_path = 'adcprep'
        else:
            if not isinstance(adcprep_path, Path):
                adcprep_path = Path(adcprep_path)
            adcprep_path = adcprep_path.as_posix()

        if aswip_command is not None:
            if isinstance(aswip_command, str):
                aswip_command = AswipCommand.from_string(aswip_command)

        self.adcprep_path = adcprep_path
        self.aswip_command = aswip_command

        setup_commands = []

        adcprep_commands = [
            f'{self.adcprep_path} --np {self.adcirc_partitions} --partmesh',
            f'{self.adcprep_path} --np {self.adcirc_partitions} --prepall',
        ]
        if self.launcher is not None:
            adcprep_commands = [f'{self.launcher} {line}' for line in adcprep_commands]
        setup_commands.extend(adcprep_commands)

        if self.aswip_command is not None:
            setup_commands.extend(
                ['', '## make sure ATCF format is correct for GAHM', f'{self.aswip_command}']
            )

        self.commands.extend(setup_commands)


class AdcircEnsembleGenerationJob(EnsembleGenerationJob):
    """
    job script to generate the ensemble configuration
    """

    def __init__(
        self,
        platform: Platform,
        slurm_tasks: int = None,
        slurm_duration: timedelta = None,
        slurm_account: str = None,
        commands: List[str] = None,
        parallel: bool = False,
        **kwargs,
    ):
        super().__init__(
            platform=platform,
            generate_command='generate_adcirc',
            commands=commands,
            slurm_run_name='ADCIRC_GENERATE_CONFIGURATION',
            slurm_tasks=slurm_tasks,
            slurm_duration=slurm_duration,
            slurm_account=slurm_account,
            parallel=parallel,
            **kwargs,
        )


class AdcircEnsembleRunScript(EnsembleRunScript):
    """
    script to run the ensemble, either by running it directly or by submitting model execution to the job manager
    default filename is ``run_<platform>.sh``
    """

    def _spinup_lines(self):
        spinup_lines = ['# run spinup', 'pushd ${DIRECTORY}/spinup >/dev/null 2>&1']
        if self.platform.value['uses_slurm']:
            dependencies = ['$setup_jobid']
            if len(dependencies) > 0:
                dependencies = f'--dependency=afterok:{":".join(dependencies)}'
            else:
                dependencies = ''
            # NOTE: `sbatch` will only use `--dependency` if it is BEFORE the job filename
            spinup_lines.extend(
                [
                    "setup_jobid=$(sbatch setup.job | awk '{print $NF}')",
                    f"spinup_jobid=$(sbatch {dependencies} adcirc.job | awk '{{print $NF}}')",
                ]
            )
        else:
            spinup_lines.extend(['sh setup.job', 'sh adcirc.job'])
        spinup_lines.extend(['popd >/dev/null 2>&1', ''])

        return spinup_lines

    def _hotstart_lines(self):
        hotstart_lines = ['pushd ${hotstart} >/dev/null 2>&1']
        if self.platform.value['uses_slurm']:
            dependencies = ['$setup_jobid']
            if self.run_spinup:
                dependencies.append('$spinup_jobid')
            if len(dependencies) > 0:
                dependencies = f'--dependency=afterok:{":".join(dependencies)}'
            else:
                dependencies = ''
            # NOTE: `sbatch` will only use `--dependency` if it is BEFORE the job filename
            hotstart_lines.extend(
                [
                    f"setup_jobid=$(sbatch setup.job | awk '{{print $NF}}')",
                    f'sbatch {dependencies} adcirc.job',
                ]
            )
        else:
            hotstart_lines.extend(['sh setup.job', 'sh adcirc.job'])
        hotstart_lines.append('popd >/dev/null 2>&1')

        return hotstart_lines


class AdcircEnsembleCleanupScript(EnsembleCleanupScript):
    """
    script for cleaning an ensemble configuration, by deleting output and log files
    """

    def __init__(self, commands: List[str] = None):
        filenames = [
            'PE*',
            'ADC_*',
            'max*',
            'partmesh.txt',
            'metis_graph.txt',
            'fort.16',
            'fort.80',
        ]
        spinup_filenames = ['fort.6*']
        hotstart_filenames = ['fort.61*', 'fort.62*', 'fort.63*', 'fort.64*']

        super().__init__(commands, filenames, spinup_filenames, hotstart_filenames)



from concurrent.futures import ProcessPoolExecutor
import logging
import os
from os import PathLike
from pathlib import Path
import shutil
import sys
import traceback

import numpy
from pyproj import CRS, Geod, Transformer
from shapely.geometry import Point


def repository_root(path: PathLike = None) -> Path:
    if path is None:
        path = __file__
    if not isinstance(path, Path):
        path = Path(path)
    if path.is_file():
        path = path.parent
    if '.git' in (child.name for child in path.iterdir()) or path == path.parent:
        return path
    else:
        return repository_root(path.parent)


def get_logger(
    name: str,
    log_filename: PathLike = None,
    file_level: int = None,
    console_level: int = None,
    log_format: str = None,
) -> logging.Logger:
    """
    instantiate logger instance

    :param name: name of logger
    :param log_filename: path to log file
    :param file_level: minimum log level to write to log file
    :param console_level: minimum log level to print to console
    :param log_format: logger message format
    :return: instance of a Logger object
    """

    if file_level is None:
        file_level = logging.DEBUG
    if console_level is None:
        console_level = logging.INFO
    logger = logging.getLogger(name)

    # check if logger is already configured
    if logger.level == logging.NOTSET and len(logger.handlers) == 0:
        # check if logger has a parent
        if '.' in name:
            if isinstance(logger.parent, logging.RootLogger):
                for existing_console_handler in [
                    handler
                    for handler in logger.parent.handlers
                    if not isinstance(handler, logging.FileHandler)
                ]:
                    logger.parent.removeHandler(existing_console_handler)
            logger.parent = get_logger(name.rsplit('.', 1)[0])
        else:
            # otherwise create a new split-console logger
            if console_level != logging.NOTSET:
                for existing_console_handler in [
                    handler
                    for handler in logger.handlers
                    if not isinstance(handler, logging.FileHandler)
                ]:
                    logger.removeHandler(existing_console_handler)

                console_output = logging.StreamHandler(sys.stdout)
                console_output.setLevel(console_level)
                logger.addHandler(console_output)

    if log_filename is not None:
        file_handler = logging.FileHandler(log_filename)
        file_handler.setLevel(file_level)
        for existing_file_handler in [
            handler for handler in logger.handlers if isinstance(handler, logging.FileHandler)
        ]:
            logger.removeHandler(existing_file_handler)
        logger.addHandler(file_handler)

    if log_format is None:
        log_format = '[%(asctime)s] %(name)-15s %(levelname)-8s: %(message)s'
    log_formatter = logging.Formatter(log_format)
    for handler in logger.handlers:
        handler.setFormatter(log_formatter)

    return logger


LOGGER = get_logger('cplmdldrv')


def create_symlink(
    source_filename: PathLike, symlink_filename: PathLike, relative: bool = False
):
    if not isinstance(source_filename, Path):
        source_filename = Path(source_filename)
    if not isinstance(symlink_filename, Path):
        symlink_filename = Path(symlink_filename)

    if not symlink_filename.parent.exists():
        symlink_filename.parent.mkdir(parents=True, exist_ok=True)

    if symlink_filename.is_symlink():
        LOGGER.debug(f'removing symlink "{symlink_filename}"')
        os.remove(symlink_filename)
    symlink_filename = symlink_filename.parent.absolute().resolve() / symlink_filename.name

    starting_directory = None
    if relative:
        starting_directory = Path().cwd().resolve()
        os.chdir(symlink_filename.parent)
        if os.path.isabs(source_filename):
            try:
                source_filename = Path(
                    os.path.relpath(source_filename, symlink_filename.parent)
                )
            except ValueError as error:
                LOGGER.warning(error)
                os.chdir(starting_directory)
    else:
        source_filename = source_filename.absolute()

    try:
        symlink_filename.symlink_to(source_filename)
    except Exception as error:
        LOGGER.warning(f'could not create symbolic link: {error}')
        shutil.copyfile(source_filename, symlink_filename)
    finally:
        if starting_directory is not None:
            os.chdir(starting_directory)


def ellipsoidal_distance(
    point_a: (float, float), point_b: (float, float), crs_a: CRS, crs_b: CRS = None
) -> float:
    if isinstance(point_a, Point):
        point_a = [*point_a.coords]
    if isinstance(point_b, Point):
        point_b = [*point_b.coords]
    if crs_b is not None:
        transformer = Transformer.from_crs(crs_b, crs_a)
        point_b = transformer.transform(*point_b)
    points = numpy.stack((point_a, point_b), axis=0)
    ellipsoid = crs_a.datum.to_json_dict()['ellipsoid']
    geodetic = Geod(a=ellipsoid['semi_major_axis'], rf=ellipsoid['inverse_flattening'])
    return geodetic.line_length(points[:, 0], points[:, 1])


def make_executable(path: PathLike):
    """
    https://stackoverflow.com/questions/12791997/how-do-you-do-a-simple-chmod-x-from-within-python
    """
    mode = os.stat(path).st_mode
    mode |= (mode & 0o444) >> 2  # copy R bits to X
    os.chmod(path, mode)


class ProcessPoolExecutorStackTraced(ProcessPoolExecutor):
    """
    preserves the traceback of any kind of raised exception
    """

    def submit(self, fn, *args, **kwargs):
        return super(ProcessPoolExecutorStackTraced, self).submit(
            self._function_wrapper, fn, *args, **kwargs,
        )

    @staticmethod
    def _function_wrapper(fn, *args, **kwargs):
        try:
            return fn(*args, **kwargs)
        except Exception:
            # Creates an exception of the same type with the traceback as message
            raise sys.exc_info()[0](traceback.format_exc())



from datetime import datetime, timedelta
from os import PathLike
from pathlib import Path
from typing import Any, Dict, List, Union

from adcircpy import AdcircMesh, AdcircRun, Tides
from adcircpy.forcing import BestTrackForcing
from adcircpy.forcing.base import Forcing
from adcircpy.mesh.fort13 import NodalAttributes
from adcircpy.mesh.mesh import ModelForcings
from adcircpy.server import SlurmConfig
from nemspy.model import ADCIRCEntry

from coupledmodeldriver.configure import Model, ModelJSON, SlurmJSON
from coupledmodeldriver.configure.base import AttributeJSON, NEMSCapJSON
from coupledmodeldriver.configure.configure import from_user_input
from coupledmodeldriver.configure.forcings.base import ForcingJSON
from coupledmodeldriver.utilities import LOGGER

ADCIRCPY_ATTRIBUTES = [
    'predictor_corrector',
    'RUNDES',
    '_IHOT',
    'RUNID',
    'NFOVER',
    'WarnElev',
    'iWarnElevDump',
    'WarnElevDumpLimit',
    'ErrorElev',
    'NABOUT',
    'NSCREEN',
    'IDEN',
    'NOLIBF',
    'NOLICA',
    'NOLICAT',
    'NCOR',
    'NTIP',
    'G',
    'TAU0',
    'DTDP',
    'STATIM',
    'REFTIM',
    'DRAMP',
    'DRAMPExtFlux',
    'FluxSettlingTime',
    'DRAMPIntFlux',
    'DRAMPElev',
    'DRAMPTip',
    'DRAMPMete',
    'DRAMPWRad',
    'DUnRampMete',
    'H0',
    'NODEDRYMIN',
    'NODEWETRMP',
    'VELMIN',
    'SLAM0',
    'SFEA0',
    'FFACTOR',
    'CF',
    'HBREAK',
    'FTHETA',
    'FGAMMA',
    'ESLM',
    'NOUTGE',
    'TOUTSGE',
    'TOUTFGE',
    'NSPOOLGE',
    'NOUTGV',
    'TOUTSGV',
    'TOUTFGV',
    'NSPOOLGV',
    'NOUTGM',
    'TOUTSGM',
    'TOUTFGM',
    'NSPOOLGM',
    'NOUTGC',
    'TOUTSGC',
    'TOUTFGC',
    'NSPOOLGC',
    'CORI',
    'ANGINN',
    'THAS',
    'THAF',
    'NHAINC',
    'FMV',
    'NHSTAR',
    'NHSINC',
    'ITITER',
    'ISLDIA',
    'CONVCR',
    'ITMAX',
    'NCPROJ',
    'NCINST',
    'NCSOUR',
    'NCHIST',
    'NCREF',
    'NCCOM',
    'NCHOST',
    'NCCONV',
    'NCCOUT',
    'vertical_mode',
    'lateral_stress_in_gwce',
    'lateral_stress_in_gwce_is_symmetrical',
    'adcvection_in_gwce',
    'lateral_stress_in_momentum',
    'lateral_stress_in_momentum_is_symmetrical',
    'lateral_stress_in_momentum_method',
    'adcvection_in_momentum',
    'area_integration_in_momentum',
    'baroclinicity',
    'gwce_solution_scheme',
    'passive_scalar_transport',
    'stress_based_3D',
    'smagorinsky',
    'smagorinsky_coefficient',
    'horizontal_mixing_coefficient',
    'CFL',
    'ICS',
]
OUTPUT_INTERVAL_DEFAULTS = {
    'surface_output_interval': timedelta(hours=1),
    'stations_output_interval': timedelta(minutes=6),
}


class ADCIRCJSON(ModelJSON, NEMSCapJSON, AttributeJSON):
    """
    ADCIRC configuration in ``configure_adcirc.json``

    stores a number of ADCIRC parameters (``ICS``, ``IM``, etc.) and optionally NEMS parameters

    .. code-block:: python

        configuration = ADCIRCJSON(
            adcirc_executable_path='adcirc',
            adcprep_executable_path='adcprep',
            modeled_start_time=datetime(2012, 10, 22, 6),
            modeled_end_time=datetime(2012, 10, 22, 6) + timedelta(days=14.5),
            modeled_timestep=timedelta(seconds=2),
            fort_13_path=None,
            fort_14_path=INPUT_DIRECTORY / 'meshes' / 'shinnecock' / 'fort.14',
            tidal_spinup_duration=timedelta(days=12.5),
        )

    """

    name = 'ADCIRC'
    default_filename = f'configure_adcirc.json'
    default_processors = 11
    default_attributes = ADCIRCPY_ATTRIBUTES

    field_types = {
        'adcirc_executable_path': Path,
        'adcprep_executable_path': Path,
        'aswip_executable_path': Path,
        'modeled_start_time': datetime,
        'modeled_end_time': datetime,
        'modeled_timestep': timedelta,
        'fort_13_path': Path,
        'fort_14_path': Path,
        'tidal_spinup_duration': timedelta,
        'tidal_spinup_timestep': timedelta,
        'source_filename': Path,
        'use_original_mesh': bool,
        'output_surface': bool,
        'surface_output_interval': timedelta,
        'output_stations': bool,
        'stations_file_path': Path,
        'stations_output_interval': timedelta,
        'output_spinup': bool,
        'output_elevations': bool,
        'output_velocities': bool,
        'output_concentrations': bool,
        'output_meteorological_factors': bool,
    }

    def __init__(
        self,
        adcirc_executable_path: PathLike,
        adcprep_executable_path: PathLike,
        modeled_start_time: datetime,
        modeled_end_time: datetime,
        modeled_timestep: timedelta,
        fort_13_path: PathLike,
        fort_14_path: PathLike,
        tidal_spinup_duration: timedelta = None,
        tidal_spinup_timestep: timedelta = None,
        forcings: List[Forcing] = None,
        aswip_executable_path: PathLike = None,
        source_filename: PathLike = None,
        slurm_configuration: SlurmJSON = None,
        use_original_mesh: bool = False,
        output_surface: bool = True,
        surface_output_interval: timedelta = None,
        output_stations: bool = False,
        stations_output_interval=None,
        stations_file_path: PathLike = None,
        output_spinup: bool = True,
        output_elevations: bool = True,
        output_velocities: bool = True,
        output_concentrations: bool = False,
        output_meteorological_factors: bool = False,
        processors: int = 11,
        nems_parameters: Dict[str, str] = None,
        attributes: Dict[str, Any] = None,
        **kwargs,
    ):
        """
        :param adcirc_executable_path: file path to ``adcirc`` or ``NEMS.x``
        :param adcprep_executable_path: file path to ``adcprep``
        :param aswip_executable_path: file path to ``aswip``
        :param modeled_start_time: start time in model run
        :param modeled_end_time: edn time in model run
        :param modeled_timestep: time interval between model steps
        :param fort_13_path: file path to ``fort.13``
        :param fort_14_path: file path to ``fort.14``
        :param tidal_spinup_duration: tidal spinup duration for ADCIRC coldstart
        :param tidal_spinup_timestep: tidal spinup modeled time interval for ADCIRC coldstart
        :param forcings: list of Forcing objects to apply to the mesh
        :param source_filename: path to modulefile to ``source``
        :param slurm_configuration: Slurm configuration object
        :param use_original_mesh: whether to symlink / copy original mesh instead of rewriting with ``adcircpy``
        :param output_surface: write surface (entire mesh) to NetCDF
        :param surface_output_interval: frequency at which output is written to file
        :param output_stations: write stations to NetCDF (only applicable if stations file exists)
        :param stations_output_interval: frequency at which stations output is written to file
        :param stations_file_path: file path to stations file
        :param output_spinup: write spinup to NetCDF
        :param output_elevations: write elevations to NetCDF
        :param output_velocities: write velocities to NetCDF
        :param output_concentrations: write concentrations to NetCDF
        :param output_meteorological_factors: write meteorological factors to NetCDF
        :param processors: number of processors to use
        :param nems_parameters: parameters to give to NEMS cap
        :param attributes: attributes to set in ``adcircpy.AdcircRun`` object
        """

        self.__mesh = None
        self.__base_mesh = None

        if tidal_spinup_timestep is None:
            tidal_spinup_timestep = modeled_timestep

        if 'fields' not in kwargs:
            kwargs['fields'] = {}
        kwargs['fields'].update(ADCIRCJSON.field_types)

        ModelJSON.__init__(self, model=Model.ADCIRC, **kwargs)
        NEMSCapJSON.__init__(
            self, processors=processors, nems_parameters=nems_parameters, **kwargs
        )
        AttributeJSON.__init__(self, attributes=attributes, **kwargs)

        self['adcirc_executable_path'] = adcirc_executable_path
        self['adcprep_executable_path'] = adcprep_executable_path
        self['aswip_executable_path'] = aswip_executable_path
        self['modeled_start_time'] = modeled_start_time
        self['modeled_end_time'] = modeled_end_time
        self['modeled_timestep'] = modeled_timestep
        self['fort_13_path'] = fort_13_path
        self['fort_14_path'] = fort_14_path
        self['tidal_spinup_duration'] = tidal_spinup_duration
        self['tidal_spinup_timestep'] = tidal_spinup_timestep
        self['source_filename'] = source_filename
        self['use_original_mesh'] = use_original_mesh

        self['output_surface'] = output_surface
        self['surface_output_interval'] = surface_output_interval
        self['output_stations'] = output_stations
        self['stations_output_interval'] = stations_output_interval
        self['stations_file_path'] = stations_file_path
        self['output_spinup'] = output_spinup
        self['output_elevations'] = output_elevations
        self['output_velocities'] = output_velocities
        self['output_concentrations'] = output_concentrations
        self['output_meteorological_factors'] = output_meteorological_factors

        self.__forcings = []
        self.forcings = forcings
        self.slurm_configuration = slurm_configuration

        for output_interval_entry, default_interval in OUTPUT_INTERVAL_DEFAULTS.items():
            if self[output_interval_entry] is None:
                LOGGER.debug(f'setting `{output_interval_entry}` to {default_interval}')
                self[output_interval_entry] = default_interval

    @property
    def forcings(self) -> List[ForcingJSON]:
        return list(self.__forcings)

    @forcings.setter
    def forcings(self, forcings: List[ForcingJSON]):
        if forcings is None:
            forcings = []
        for forcing in forcings:
            self.add_forcing(forcing)

    def add_forcing(self, forcing: ForcingJSON):
        if not isinstance(forcing, ForcingJSON):
            forcing = from_user_input(forcing)
        adcircpy_forcing = forcing.adcircpy_forcing

        existing_forcings = [
            existing_forcing.__class__.__name__ for existing_forcing in self.adcircpy_forcings
        ]
        if adcircpy_forcing.__class__.__name__ in existing_forcings:
            existing_index = existing_forcings.index(adcircpy_forcing.__class__.__name__)
        else:
            existing_index = -1
        if existing_index > -1:
            self.__forcings[existing_index] = forcing
        else:
            self.__forcings.append(forcing)

    @property
    def slurm_configuration(self) -> List[SlurmJSON]:
        return self.__slurm_configuration

    @slurm_configuration.setter
    def slurm_configuration(self, slurm_configuration: SlurmJSON):
        if isinstance(slurm_configuration, SlurmConfig):
            SlurmJSON.from_adcircpy(slurm_configuration)
        self.__slurm_configuration = slurm_configuration

    @property
    def adcircpy_forcings(self) -> List[Forcing]:
        return [forcing.adcircpy_forcing for forcing in self.forcings]

    @property
    def adcircpy_mesh(self) -> AdcircMesh:
        if self.__mesh is None:
            mesh = self.base_mesh.copy()

            # reconstruct mesh from base
            mesh.forcings = ModelForcings(mesh)
            mesh.nodal_attributes = NodalAttributes(mesh)

            if self['fort_13_path'] is not None:
                if self['fort_13_path'].exists():
                    LOGGER.info(f'reading attributes from "{self["fort_13_path"]}"')
                    mesh.import_nodal_attributes(self['fort_13_path'])
                    for attribute_name in mesh.get_nodal_attribute_names():
                        mesh.set_nodal_attribute_state(
                            attribute_name, coldstart=True, hotstart=True
                        )
                else:
                    LOGGER.warning(
                        f'mesh values (nodal attributes) not found at "{self["fort_13_path"]}"'
                    )

            LOGGER.debug(f'adding {len(self.forcings)} forcing(s) to mesh')
            for adcircpy_forcing in self.adcircpy_forcings:
                if isinstance(adcircpy_forcing, (Tides, BestTrackForcing)):
                    adcircpy_forcing.start_date = self['modeled_start_time']
                    adcircpy_forcing.end_date = self['modeled_end_time']

                if (
                    isinstance(adcircpy_forcing, Tides)
                    and self['tidal_spinup_duration'] is not None
                ):
                    adcircpy_forcing.spinup_time = self['tidal_spinup_duration']
                    adcircpy_forcing.start_date -= self['tidal_spinup_duration']
                # elif isinstance(adcircpy_forcing, BestTrackForcing):
                #     adcircpy_forcing.clip_to_bbox(mesh.get_bbox(output_type='bbox'), mesh.crs)

                mesh.add_forcing(adcircpy_forcing)

            if not mesh.has_nodal_attribute('primitive_weighting_in_continuity_equation'):
                LOGGER.debug(f'generating tau0 in mesh')
                mesh.generate_tau0()

            self.__mesh = mesh

        return self.__mesh

    @adcircpy_mesh.setter
    def adcircpy_mesh(self, adcircpy_mesh: Union[AdcircMesh, PathLike]):
        self.__mesh = adcircpy_mesh

    @property
    def base_mesh(self) -> AdcircMesh:
        if self.__base_mesh is None:
            if self.__mesh is not None:
                self.__base_mesh = self.__mesh
            else:
                self.__base_mesh = self['fort_14_path']
        if not isinstance(self.__base_mesh, AdcircMesh):
            LOGGER.info(f'opening mesh "{self.__base_mesh}"')
            self.__base_mesh = AdcircMesh.open(self.__base_mesh, crs=4326)
        # deconstruct mesh into a base mesh that can be pickled
        self.__base_mesh.forcings = None
        self.__base_mesh.nodal_attributes = None
        return self.__base_mesh

    @base_mesh.setter
    def base_mesh(self, base_mesh: Union[AdcircMesh, PathLike]):
        self.__base_mesh = base_mesh

    @property
    def adcircpy_driver(self) -> AdcircRun:
        # instantiate AdcircRun object.
        driver = AdcircRun(
            mesh=self.adcircpy_mesh,
            start_date=self['modeled_start_time'],
            end_date=self['modeled_end_time'],
            spinup_time=self['tidal_spinup_duration'],
            server_config=self.slurm_configuration.to_adcircpy()
            if self.slurm_configuration is not None
            else None,
        )

        if self['stations_file_path'] is not None:
            LOGGER.info(f'importing stations from "{self["stations_file_path"]}"')
            driver.import_stations(self['stations_file_path'])

        if self['modeled_timestep'] is not None:
            driver.timestep = self['modeled_timestep'] / timedelta(seconds=1)

        for name, value in self['attributes'].items():
            if value is not None:
                try:
                    setattr(driver, name, value)
                except:
                    LOGGER.warning(
                        f'could not set `{driver.__class__.__name__}` attribute `{name}` to `{value}`'
                    )

        if self['tidal_spinup_duration'] is not None and self['output_spinup']:
            spinup_start = self['modeled_start_time'] - self['tidal_spinup_duration']
            spinup_end = self['modeled_start_time']
            spinup_output_interval = self['surface_output_interval']
        else:
            spinup_start = None
            spinup_end = None
            spinup_output_interval = None

        if self['output_elevations']:
            if self['output_surface']:
                driver.set_elevation_surface_output(
                    sampling_rate=self['surface_output_interval'],
                    start=self['modeled_start_time'],
                    end=self['modeled_end_time'],
                    spinup=spinup_output_interval,
                    spinup_start=spinup_start,
                    spinup_end=spinup_end,
                )
            if self['output_stations']:
                driver.set_elevation_stations_output(
                    sampling_rate=self['stations_output_interval'],
                    start=self['modeled_start_time'],
                    end=self['modeled_end_time'],
                    spinup=spinup_output_interval,
                    spinup_start=spinup_start,
                    spinup_end=spinup_end,
                )

        if self['output_velocities']:
            if self['output_surface']:
                driver.set_velocity_surface_output(
                    sampling_rate=self['surface_output_interval'],
                    start=self['modeled_start_time'],
                    end=self['modeled_end_time'],
                    spinup=spinup_output_interval,
                    spinup_start=spinup_start,
                    spinup_end=spinup_end,
                )
            if self['output_stations']:
                driver.set_velocity_stations_output(
                    sampling_rate=self['stations_output_interval'],
                    start=self['modeled_start_time'],
                    end=self['modeled_end_time'],
                    spinup=spinup_output_interval,
                    spinup_start=spinup_start,
                    spinup_end=spinup_end,
                )

        if self['output_concentrations']:
            if self['output_surface']:
                driver.set_concentration_surface_output(
                    sampling_rate=self['surface_output_interval'],
                    start=self['modeled_start_time'],
                    end=self['modeled_end_time'],
                    spinup=spinup_output_interval,
                    spinup_start=spinup_start,
                    spinup_end=spinup_end,
                )
            if self['output_stations']:
                driver.set_concentration_stations_output(
                    sampling_rate=self['stations_output_interval'],
                    start=self['modeled_start_time'],
                    end=self['modeled_end_time'],
                    spinup=spinup_output_interval,
                    spinup_start=spinup_start,
                    spinup_end=spinup_end,
                )

        if self['output_meteorological_factors']:
            if self['output_surface']:
                driver.set_meteorological_surface_output(
                    sampling_rate=self['surface_output_interval'],
                    start=self['modeled_start_time'],
                    end=self['modeled_end_time'],
                    spinup=spinup_output_interval,
                    spinup_start=spinup_start,
                    spinup_end=spinup_end,
                )
            if self['output_stations']:
                driver.set_meteorological_stations_output(
                    sampling_rate=self['stations_output_interval'],
                    start=self['modeled_start_time'],
                    end=self['modeled_end_time'],
                    spinup=spinup_output_interval,
                    spinup_start=spinup_start,
                    spinup_end=spinup_end,
                )

        return driver

    @property
    def nemspy_entry(self) -> ADCIRCEntry:
        return ADCIRCEntry(processors=self['processors'], **self['nems_parameters'])

    def __setitem__(self, key: str, value: Any):
        super().__setitem__(key, value)
        self.__mesh = None

    def __copy__(self) -> 'ADCIRCJSON':
        instance = super().__copy__()
        instance.base_mesh = self.base_mesh
        instance.forcings = self.forcings
        return instance



from coupledmodeldriver.configure.forcings.base import (
    ATMESHForcingJSON,
    BestTrackForcingJSON,
    OWIForcingJSON,
    TidalForcingJSON,
    WW3DATAForcingJSON,
    NationalWaterModelFocringJSON,
)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['coupledmodeldriver',
 'coupledmodeldriver.client',
 'coupledmodeldriver.configure',
 'coupledmodeldriver.configure.forcings',
 'coupledmodeldriver.generate',
 'coupledmodeldriver.generate.adcirc',
 'coupledmodeldriver.generate.schism']

package_data = \
{'': ['*']}

install_requires = \
['adcircpy>=1.2.1,<2.0.0',
 'file-read-backwards',
 'nemspy>=1.0.4',
 'numpy',
 'pyproj',
 'pyschism',
 'typepigeon>=1.0.3,<2.0.0']

extras_require = \
{'development': ['isort', 'oitnb'],
 'documentation': ['m2r2',
                   'sphinx',
                   'sphinx-rtd-theme',
                   'sphinxcontrib-programoutput'],
 'testing': ['pooch', 'pytest', 'pytest-cov', 'pytest-socket', 'pytest-xdist']}

entry_points = \
{'console_scripts': ['check_completion = '
                     'coupledmodeldriver.client.check_completion:main',
                     'generate_adcirc = '
                     'coupledmodeldriver.client.generate_adcirc:main',
                     'generate_schism = '
                     'coupledmodeldriver.client.generate_schism:main',
                     'initialize_adcirc = '
                     'coupledmodeldriver.client.initialize_adcirc:main',
                     'initialize_schism = '
                     'coupledmodeldriver.client.initialize_schism:main',
                     'unqueued_runs = '
                     'coupledmodeldriver.client.unqueued_runs:main']}

setup_kwargs = {
    'name': 'coupledmodeldriver',
    'version': '1.6.1',
    'description': 'coupled model configuration generation',
    'long_description': '# CoupledModelDriver\n\n[![tests](https://github.com/noaa-ocs-modeling/CoupledModelDriver/workflows/tests/badge.svg)](https://github.com/noaa-ocs-modeling/CoupledModelDriver/actions?query=workflow%3Atests)\n[![codecov](https://codecov.io/gh/noaa-ocs-modeling/coupledmodeldriver/branch/main/graph/badge.svg?token=4DwZePHp18)](https://codecov.io/gh/noaa-ocs-modeling/coupledmodeldriver)\n[![build](https://github.com/noaa-ocs-modeling/CoupledModelDriver/workflows/build/badge.svg)](https://github.com/noaa-ocs-modeling/CoupledModelDriver/actions?query=workflow%3Abuild)\n[![version](https://img.shields.io/pypi/v/CoupledModelDriver)](https://pypi.org/project/CoupledModelDriver)\n[![license](https://img.shields.io/github/license/noaa-ocs-modeling/CoupledModelDriver)](https://creativecommons.org/share-your-work/public-domain/cc0)\n[![style](https://sourceforge.net/p/oitnb/code/ci/default/tree/_doc/_static/oitnb.svg?format=raw)](https://sourceforge.net/p/oitnb/code)\n[![documentation](https://readthedocs.org/projects/coupledmodeldriver/badge/?version=latest)](https://coupledmodeldriver.readthedocs.io/en/latest/?badge=latest)\n\nCoupledModelDriver generates an overlying job submission framework and configuration directories for NEMS-coupled coastal ocean\nmodel ensembles.\n\n```shell\npip install coupledmodeldriver\n```\n\nIt utilizes [NEMSpy](https://nemspy.readthedocs.io) to generate NEMS configuration files, shares common configurations between\nruns, and organizes spinup and mesh partition into separate jobs for dependant submission.\n\nDocumentation can be found at https://coupledmodeldriver.readthedocs.io\n\n## supported models and platforms\n\n- **models**\n    - circulation models\n        - ADCIRC (uses [ADCIRCpy](https://pypi.org/project/adcircpy))\n        - SCHISM (uses [PySCHISM](https://github.com/schism-dev/pyschism))\n    - forcings\n        - ATMESH\n        - WW3DATA\n        - HURDAT best track\n        - OWI\n- **platforms**\n    - local\n    - Slurm\n        - Hera\n        - Stampede2\n        - Orion\n\n## organization / responsibility\n\nCoupledModelDriver is developed for the COASTAL Act project by the [Coastal Marine Modeling Branch (CMMB)](https://coastaloceanmodels.noaa.gov) of the Office of Coast Survey (OCS), a part of the [National Oceanic and Atmospheric Administration (NOAA)](https://www.noaa.gov), an agency of the United States federal government.\n\n- Zachary Burnett (**lead**) - zachary.burnett@noaa.gov\n- William Pringle - wpringle@anl.gov\n- Saeed Moghimi - saeed.moghimi@noaa.gov\n\n## usage example\n\n### 1. generate JSON configuration files\n\n`initialize_adcirc` creates JSON configuration files according to the given parameters. ADCIRC run options that are not exposed\nby this command, such as `runs` or `gwce_solution_scheme`, can be specified by directly modifying the JSON files. The following\ncreates JSON files for coupling `(ATMESH + WW3DATA) -> ADCIRC` over a small Shinnecock Inlet mesh:\n\n```shell\ninitialize_adcirc \\\n    --platform HERA \\\n    --mesh-directory /scratch2/COASTAL/coastal/save/shared/models/meshes/shinnecock/v1.0 \\\n    --output-directory hera_shinnecock_ike_spinup_tidal_atmesh_ww3data \\\n    --modeled-start-time 20080823 \\\n    --modeled-duration 14:06:00:00 \\\n    --modeled-timestep 00:00:02 \\\n    --nems-interval 01:00:00 \\\n    --adcirc-executable /scratch2/COASTAL/coastal/save/shared/repositories/CoastalApp/ALLBIN_INSTALL/NEMS-adcirc-atmesh-ww3data.x \\\n    --adcirc-processors 40\n    --adcprep-executable /scratch2/COASTAL/coastal/save/shared/repositories/CoastalApp/ALLBIN_INSTALL/adcprep \\\n    --modulefile /scratch2/COASTAL/coastal/save/shared/repositories/CoastalApp/modulefiles/envmodules_intel.hera \\\n    --forcings tidal,atmesh,ww3data \\\n    --tidal-source TPXO \\\n    --tidal-path /scratch2/COASTAL/coastal/save/shared/models/forcings/tides/h_tpxo9.v1.nc \\\n    --tidal-spinup-duration 12:06:00:00 \\\n    --atmesh-path /scratch2/COASTAL/coastal/save/shared/models/forcings/shinnecock/ike/wind_atm_fin_ch_time_vec.nc \\\n    --ww3data-path /scratch2/COASTAL/coastal/save/shared/models/forcings/shinnecock/ike/ww3.Constant.20151214_sxy_ike_date.nc\n```\n\nThis will create the directory `hera_shinnecock_ike_spinup_tidal_atmesh_ww3data/` with the following JSON configuration files:\n\n```\n📂 hera_shinnecock_ike_spinup_tidal_atmesh_ww3data/\n┣ 📜 configure_adcirc.json\n┣ 📜 configure_atmesh.json\n┣ 📜 configure_modeldriver.json\n┣ 📜 configure_nems.json\n┣ 📜 configure_slurm.json\n┣ 📜 configure_tidal_forcing.json\n┗ 📜 configure_ww3data.json\n```\n\nThese files contain relevant configuration values for an ADCIRC run. You will likely wish to change these values to alter the\nresulting run, before generating the actual model configuration. For instance, NEMS connections and the run sequence need to be\nmanually specified in `configure_nems.json`.\n\n### 2. generate model configuration files\n\n`generate_adcirc` generates an ADCIRC run configuration (`fort.14`, `fort.15`, etc.) using options read from the JSON\nconfiguration files (generated in the previous step).\n\n```shell\ncd hera_shinnecock_ike_spinup_tidal_atmesh_ww3data\ngenerate_adcirc\n```\n\nThe resulting configuration will look like this:\n\n```\n📂 hera_shinnecock_ike_spinup_tidal_atmesh_ww3data/\n┣ 📜 configure_adcirc.json\n┣ 📜 configure_atmesh.json\n┣ 📜 configure_modeldriver.json\n┣ 📜 configure_nems.json\n┣ 📜 configure_slurm.json\n┣ 📜 configure_tidal_forcing.json\n┣ 📜 configure_ww3data.json\n┣ 📂 spinup/\n┃  ┣ 📜 fort.13\n┃  ┣ 🔗 fort.14 -> ../fort.14\n┃  ┣ 📜 fort.15\n┃  ┣ 📜 nems.configure\n┃  ┣ 📜 model_configure\n┃  ┣ 🔗 atm_namelist.rc -> ./model_configure\n┃  ┣ 📜 config.rc\n┃  ┣ 📜 setup.job\n┃  ┗ 📜 adcirc.job\n┣ 📂 runs/\n┃  ┗ 📂 unperturbed/\n┃    ┣ 📜 fort.13\n┃    ┣ 🔗 fort.14 -> ../../fort.14\n┃    ┣ 📜 fort.15\n┃    ┣ 🔗 fort.67.nc -> ../../spinup/fort.67.nc\n┃    ┣ 🔗 fort.68.nc -> ../../spinup/fort.68.nc\n┃    ┣ 📜 nems.configure\n┃    ┣ 📜 model_configure\n┃    ┣ 🔗 atm_namelist.rc -> ./model_configure\n┃    ┣ 📜 config.rc\n┃    ┣ 📜 setup.job\n┃    ┗ 📜 adcirc.job\n┣ 📜 fort.14\n┣ 📜 cleanup.sh\n┗ 📜 run_hera.sh\n```\n\n### 3. run the model\n\nThe previous step will also have generated a script called `./run_hera.sh`. You can run it to submit the model run to the Slurm\njob queue:\n\n```shell\n./run_hera.sh\n``` \n\nThe queue will have the following jobs added:\n\n```\n   JOBID CPU NODE DEPENDENCY       NODELIST(REA NAME\n20967647 1   1    (null)           (None)       ADCIRC_SETUP_SPINUP\n20967648 40  1    afterok:20967647 (Dependency) ADCIRC_COLDSTART_SPINUP\n20967649 1   1    (null)           (None)       ADCIRC_SETUP_unperturbed\n20967650 42  2    afterok:20967649 (Dependency) ADCIRC_HOTSTART_unperturbed\n```\n\n### 4. track model progress\n\n`check_completion` checks the completion status of a running model directory.\n\n```shell\ncd hera_shinnecock_ike_spinup_tidal_atmesh_ww3data\ncheck_completion\n```\n\n```json\n{\n    "hera_shinnecock_ike_spinup_tidal_atmesh_ww3data": {\n        "spinup": "running - 15%",\n        "runs": "not_started - 0%"\n    }\n}\n```\n\nyou can also pass a specific directory (or several directories):\n\n```shell\ncheck_completion spinup\n```\n\n```json\n{\n    "spinup": "running - 27%"\n}\n```\n\n```shell\ncd run_20211027_florence_besttrack_250msubset_quadrature\ncheck_completion runs/*_13\n```\n\n```json\n{\n    "vortex_4_variable_perturbation_13": "completed - 100.0%",\n    "vortex_4_variable_quadrature_13": "not_started - 0%"\n}\n```\n\nif a run has an error, you can pass `--verbose` to see detailed logs:\n\n```shell\ncheck_completion spinup\n```\n\n```json\n{\n    "spinup": "error - 0%"\n}\n```\n\n```shell\ncheck_completion spinup --verbose\n```\n\n```json\n{\n    "spinup": {\n        "status": "error",\n        "progress": "0%",\n        "error": {\n            "ADCIRC_SETUP_SPINUP.err.log": [\n                "forrtl: severe (24): end-of-file during read, unit -4, file /proc/92195/fd/0\\n",\n                "Image              PC                Routine            Line        Source             \\n",\n                "adcprep            000000000069A72E  Unknown               Unknown  Unknown\\n",\n                "adcprep            00000000006CBAAF  Unknown               Unknown  Unknown\\n",\n                "adcprep            000000000050A5CB  openprepfiles_           6996  prep.F\\n",\n                "adcprep            0000000000507F22  prep13_                   753  prep.F\\n",\n                "adcprep            000000000042E2E9  prepinput_                717  adcprep.F\\n",\n                "adcprep            000000000042BCDB  MAIN__                    239  adcprep.F\\n",\n                "adcprep            000000000040B65E  Unknown               Unknown  Unknown\\n",\n                "libc-2.17.so       00002AAEC02EB555  __libc_start_main     Unknown  Unknown\\n",\n                "adcprep            000000000040B569  Unknown               Unknown  Unknown\\n",\n                "srun: error: h24c51: task 0: Exited with exit code 24\\n",\n                "srun: launch/slurm: _step_signal: Terminating StepId=25366266.1\\n"\n            ]\n        }\n    }\n}\n```\n\n```shell\ncheck_completion runs\n```\n\n```json\n{\n    "spinup": "failed - 0%"\n}\n```\n\n```shell\ncheck_completion runs --verbose\n```\n\n```json\n{\n    "runs": {\n        "status": "failed",\n        "progress": "0%",\n        "failed": {\n            "fort.16": "ADCIRC output file `fort.16` not found"\n        },\n        "error": {\n            "ADCIRC_SETUP_unperturbed.err.log": [\n                "slurmstepd: error: execve(): /scratch2/COASTAL/coastal/save/shared/repositories/CoastalApp/ADCIRC/ALLBIN_INSTALL/adcprep: No such file or directory\\n",\n                "srun: error: h18c49: task 0: Exited with exit code 2\\n",\n                "srun: launch/slurm: _step_signal: Terminating StepId=25366268.0\\n"\n            ]\n        }\n    }\n}\n```\n',
    'author': 'Zach Burnett',
    'author_email': 'zachary.r.burnett@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/noaa-ocs-modeling/CoupledModelDriver.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)
